#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Calculate the optical path due to atomic and molecular transitions.
The opacity of a mixture of gases is calculated by the correlated-k method and
using pre-tabulated ktables, assuming random overlap.
"""
import numpy as np
from numba import jit

@jit(nopython=True)
def interp_k(P_grid, T_grid, P_layer, T_layer, k_gas_w_g_p_t):
    """
    Interpolate the k coeffcients to given presures and temperatures
    using pre-tabulated k-tables.

    Parameters
    ----------
    P_grid(NPRESSKTA) : ndarray
        Pressure grid on which the k-coefficients are pre-computed.
        Unit: Pa
    T_grid(NTEMPKTA) : ndarray
        Temperature grid on which the k-coefficients are pre-computed.
        Unit: Kelvin
    P_layer(NLAYER) : ndarray
        Atmospheric pressure grid.
        Unit: Pa
    T_layer(NLAYER) : ndarray
        Atmospheric temperature grid.
        Unit: Kelvin
    k_gas_w_g_p_t(NGAS,NWAVEKTA,NG,NPRESSKTA,NTEMPKTA) : ndarray
        Array storing the k-coefficients.

    Returns
    -------
    k_gas_w_g_l(NGAS,NWAVEKTA,NG,NLAYER) : ndarray
        The interpolated-to-atmosphere k-coefficients.
        Has dimension: NGAS x NWAVE x NG x NLAYER.
    Notes
    -----
    Code breaks if P_layer/T_layer is out of the range of P_grid/T_grid.
    Mainly need to worry about max(T_layer)>max(T_grid).
    No extrapolation outside of the TP grid of ktable.
    """
    NGAS, NWAVE, NG, NPRESS, NTEMP = k_gas_w_g_p_t.shape
    NLAYER = len(P_layer)
    k_gas_w_g_l = np.zeros((NGAS,NWAVE,NG,NLAYER))
    kgood = np.zeros((NGAS,NWAVE,NG,NLAYER))

    for ilayer in range(NLAYER):
        p = P_layer[ilayer]
        t = T_layer[ilayer]

        # Find pressure grid points above and below current layer pressure
        ip = np.abs(P_grid-p).argmin()
        if P_grid[ip] >= p:
            ip_high = ip
            if ip == 0:
                p = P_grid[0]
                ip_low = 0
                ip_high = 1
            else:
                ip_low = ip-1
        elif P_grid[ip]<p:
            ip_low = ip
            if ip == NPRESS-1:
                p = P_grid[NPRESS-1]
                ip_high = NPRESS-1
                ip_low = NPRESS-2
            else:
                ip_high = ip + 1

        # Find temperature grid points above and below current layer temperature
        it = np.abs(T_grid-t).argmin()
        if T_grid[it] >= t:
            it_high = it
            if it == 0:
                t = T_grid[0]
                it_low = 0
                it_high = 1
            else:
                it_low = it -1
        elif T_grid[it] < t:
            it_low = it
            if it == NTEMP-1:
                t = T_grid[-1]
                it_high = NTEMP - 1
                it_low = NTEMP -2
            else:
                it_high = it + 1

        # Set up arrays for interpolation
        lnp = np.log(p)
        lnp_low = np.log(P_grid[ip_low])
        lnp_high = np.log(P_grid[ip_high])
        t_low = T_grid[it_low]
        t_high = T_grid[it_high]

        f11 = np.zeros((NGAS,NWAVE,NG))
        f12 = np.zeros((NGAS,NWAVE,NG))
        f22 = np.zeros((NGAS,NWAVE,NG))
        f21 = np.zeros((NGAS,NWAVE,NG))

        f11[:,:,:] = k_gas_w_g_p_t[:,:,:,ip_low,it_low]
        f12[:,:,:] = k_gas_w_g_p_t[:,:,:,ip_low,it_high]
        f21[:,:,:] = k_gas_w_g_p_t[:,:,:,ip_high,it_high]
        f22[:,:,:] = k_gas_w_g_p_t[:,:,:,ip_high,it_low]

        # Bilinear interpolation
        v = (lnp-lnp_low)/(lnp_high-lnp_low)
        u = (t-t_low)/(t_high-t_low)

        igood = np.where( (f11>0.0) & (f12>0.0) & (f22>0.0) & (f21>0.0) )
        ibad = np.where( (f11<=0.0) & (f12<=0.0) & (f22<=0.0) & (f21<=0.0) )

        for index in range(len(igood[0])):
            kgood[igood[0][index],igood[1][index],igood[2][index],ilayer] \
                = (1.0-v)*(1.0-u)*np.log(f11[igood[0][index],igood[1][index],igood[2][index]]) \
                + v*(1.0-u)*np.log(f22[igood[0][index],igood[1][index],igood[2][index]]) \
                + v*u*np.log(f21[igood[0][index],igood[1][index],igood[2][index]]) \
                + (1.0-v)*u*np.log(f12[igood[0][index],igood[1][index],igood[2][index]])
            kgood[igood[0][index],igood[1][index],igood[2][index],ilayer] \
                = np.exp(kgood[igood[0][index],igood[1][index],igood[2][index],ilayer])

        for index in range(len(ibad[0])):
            kgood[ibad[0][index],ibad[1][index],ibad[2][index],ilayer] \
                = (1.0-v)*(1.0-u)*f11[ibad[0][index],ibad[1][index],ibad[2][index]] \
                + v*(1.0-u)*f22[ibad[0][index],ibad[1][index],ibad[2][index]] \
                + v*u*f21[ibad[0][index],ibad[1][index],ibad[2][index]] \
                + (1.0-v)*u*f12[ibad[0][index],ibad[1][index],ibad[2][index]]

    k_gas_w_g_l = kgood
    return k_gas_w_g_l

@jit(nopython=True)
def rankg(weight, cont, del_g):
    """
    Combine the randomly overlapped k distributions of two gases into a single
    k distribution.

    Parameters
    ----------
    weight(NG) : ndarray
        Weights of points in the random k-dist
    cont(NG) : ndarray
        Random k-coeffs in the k-dist.
    del_g(NG) : ndarray
        Required weights of final k-dist.

    Returns
    -------
    k_g(NG) : ndarray
        Combined k-dist.
        Unit: cm^2 (per particle)
    """
    ng = len(del_g)
    nloop = ng*ng

    # sum delta gs to get cumulative g ordinate
    g_ord = np.zeros(ng+1)
    g_ord[1:] = np.cumsum(del_g)
    g_ord[ng] = 1

    # Sort random k-coeffs into ascending order. Integer array ico records
    # which swaps have been made so that we can also re-order the weights.
    ico = np.argsort(cont)
    cont = cont[ico]
    weight = weight[ico] # sort weights accordingly
    gdist = np.cumsum(weight)
    k_g = np.zeros(ng)

    ig = 0
    sum1 = 0.0
    cont_weight = cont * weight
    for iloop in range(nloop):
        if gdist[iloop] < g_ord[ig+1]:
            k_g[ig] = k_g[ig] + cont_weight[iloop]
            sum1 = sum1 + weight[iloop]
        else:
            frac = (g_ord[ig+1] - gdist[iloop-1])/(gdist[iloop]-gdist[iloop-1])
            k_g[ig] = k_g[ig] + np.float32(frac)*cont_weight[iloop]

            sum1 = sum1 + frac * weight[iloop]
            k_g[ig] = k_g[ig]/np.float32(sum1)

            ig = ig +1
            sum1 = (1.0-frac)*weight[iloop]
            k_g[ig] = np.float32(1.0-frac)*cont_weight[iloop]

    if ig == ng-1:
        k_g[ig] = k_g[ig]/np.float32(sum1)

    return k_g

@jit(nopython=True)
def noverlapg(k_gas_g, amount, del_g):
    """
    Combine k distributions of multiple gases given their number densities.

    Parameters
    ----------
    k_gas_g(NGAS,NG) : ndarray
        K-distributions of the different gases.
        Each row contains a k-distribution defined at NG g-ordinates.
        Unit: cm^2 (per particle)
    amount(NGAS) : ndarray
        Absorber amount of each gas.
        Unit: (no. of partiicles) cm^-2
    del_g(NG) : ndarray
        Gauss quadrature weights for the g-ordinates.
        These are the widths of the bins in g-space.

    Returns
    -------
    k_g(NG) : ndarray
        Combined k-distribution.
        Unit: cm^2 (per particle)
    """
    # amount = VMR_layer x U_layer
    NGAS = len(amount)
    NG = len(del_g)
    k_g = np.zeros(NG)
    weight = np.zeros(NG*NG)
    contri = np.zeros(NG*NG)
    for igas in range(NGAS-1):
        # first pair of gases
        if igas == 0:
            a1 = amount[igas]
            a2 = amount[igas+1]

            k_g1 = k_gas_g[igas,:]
            k_g2 = k_gas_g[igas+1,:]

            # skip if first k-distribution = 0.0
            if k_g1[NG-1]*a1 == 0.0:
                k_g = k_g2*a2
            # skip if second k-distribution = 0.0
            elif k_g2[NG-1]*a2 == 0.0:
                k_g = k_g1*a1
            else:
                nloop = 0
                for ig in range(NG):
                    for jg in range(NG):
                        weight[nloop] = del_g[ig]*del_g[jg]
                        contri[nloop] = k_g1[ig]*a1 + k_g2[jg]*a2
                        nloop = nloop + 1
                k_g = rankg(weight,contri,del_g)
        # subsequuent gases .. add amount*k to previous summed k
        else:
            a2 = amount[igas+1]
            # print('a2',a2)
            k_g1 = k_g
            k_g2 = k_gas_g[igas+1,:]

            # skip if first k-distribution = 0.0
            if k_g1[NG-1] == 0:
                k_g = k_g2*a2
            # Skip if second k-distribution = 0.0
            elif k_g2[NG-1]*a2== 0:
                k_g = k_g1
            else:
                nloop = 0
                for ig in range(NG):
                    for jg in range(NG):
                        weight[nloop] = del_g[ig]*del_g[jg]
                        contri[nloop] = k_g1[ig]+k_g2[jg]*a2
                        nloop = nloop + 1
                k_g = rankg(weight,contri,del_g)

    return k_g

@jit(nopython=True)
def calc_tau_gas(k_gas_w_g_p_t, P_layer, T_layer, VMR_layer, U_layer,
    P_grid, T_grid, del_g):
    """
    Parameters
    ----------
    k_gas_w_g_p_t(NGAS,NWAVE,NG,NPRESSK,NTEMPK) : ndarray
        k-coefficients.
        Unit: cm^2 (per particle)
        Has dimension: NWAVE x NG x NPRESSK x NTEMPK.
    P_layer(NLAYER) : ndarray
        Atmospheric pressure grid.
        Unit: Pa
    T_layer(NLAYER) : ndarray
        Atmospheric temperature grid.
        Unit: K
    VMR_layer(NLAYER,NGAS) : ndarray
        Array of volume mixing ratios for NGAS.
        Has dimensioin: NLAYER x NGAS
    U_layer(NLAYER) : ndarray
        Total number of gas particles in each layer.
        Unit: (no. of particle) m^-2
    P_grid(NPRESSK) : ndarray
        Pressure grid on which the k-coeff's are pre-computed.
    T_grid(NTEMPK) : ndarray
        Temperature grid on which the k-coeffs are pre-computed.
    del_g(NG) : ndarray
        Gauss quadrature weights for the g-ordinates.
        These are the widths of the bins in g-space.

    Returns
    -------
    tau_w_g_l(NWAVE,NG,NLAYER) : ndarray
        Optical path due to spectral line absorptions.

    Notes
    -----
    Absorber amounts (U_layer) is scaled down by a factor 1e-20 because Nemesis
    k-tables are scaled up by a factor of 1e20.
    """

    Scaled_U_layer = U_layer * 1.0e-20
    Scaled_U_layer *= 1.0e-4 # convert from per m^2 to per cm^2

    k_gas_w_g_l = interp_k(P_grid, T_grid, P_layer, T_layer, k_gas_w_g_p_t)
    Ngas, Nwave, Ng, Nlayer = k_gas_w_g_l.shape

    amount_layer = np.zeros((Nlayer,Ngas))
    for ilayer in range(Nlayer):
        amount_layer[ilayer,:] = Scaled_U_layer[ilayer] * VMR_layer[ilayer,:4]

    tau_w_g_l = np.zeros((Nwave,Ng,Nlayer))
    for iwave in range (Nwave):
        k_gas_g_l = k_gas_w_g_l[:,iwave,:,:]
        k_g_l = np.zeros((Ng,Nlayer))
        for ilayer in range(Nlayer):
            k_g_l[:,ilayer]\
                = noverlapg(k_gas_g_l[:,:,ilayer],amount_layer[ilayer,:],del_g)
            tau_w_g_l[iwave,:,ilayer] = k_g_l[:,ilayer]

    return tau_w_g_l


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
import os
# test_path = os.path.abspath(os.path.join(os.getcwd(),os.pardir))
# root_path = os.path.abspath(os.path.join(test_path,os.pardir))
# data_path = os.path.abspath(os.path.join(root_path,'nemesispy/data'))
# ktable_path = os.path.abspath(os.path.join(data_path,'ktables'))
# cia_path = os.path.abspath(os.path.join(data_path,'cia'))

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
nemesis_location =  os.path.realpath(
    os.path.join(__location__,os.pardir)
    )

ktable_path = os.path.join(nemesis_location, "data/ktables")
lowres_file_paths = [
    'h2owasp43.kta',
    'co2wasp43.kta',
    'cowasp43.kta',
    'ch4wasp43.kta']
for ipath,path in enumerate(lowres_file_paths):
    lowres_file_paths[ipath] = os.path.join(ktable_path,path)

ariel_file_paths = [
    'H2O_Katy_ARIEL_test.kta',
    'CO2_Katy_ARIEL_test.kta',
    'CO_Katy_ARIEL_test.kta',
    'CH4_Katy_ARIEL_test.kta']
for ipath,path in enumerate(ariel_file_paths):
    ariel_file_paths[ipath] = os.path.join(ktable_path,path)

hires_file_paths = [
    'H2O_Katy_ARIEL_test.kta',
    'CO2_Katy_ARIEL_test.kta',
    'CO_Katy_ARIEL_test.kta',
    'CH4_Katy_ARIEL_test.kta']
for ipath,path in enumerate(ariel_file_paths):
    ariel_file_paths[ipath] = os.path.join(ktable_path,path)

cia_file_path = ktable_path = os.path.join(nemesis_location, "data/cia")
cia_file_path = os.path.join(cia_file_path,'exocia_hitran12_200-3800K.tab')



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Use hydrostatic equilibrium to find the altitudes of points in the atmosphere
given their pressures, temperatures and mean molecular weights.
"""
import numpy as np
from numba import jit
from nemesispy.common.constants import G, K_B

@jit(nopython=True)
def calc_grav_simple(h, M_plt, R_plt):
    """
    Calculates the gravitational acceleration at altitude h on a planet.

    Parameters
    ----------
    h : real
        Altitude.
        Unit: m
    M_plt : real
        Planet mass.
        Unit: kg
    R_plt : real
        Planet radius.
        Unit: m

    Returns
    -------
    g : real
        Gravitational acceleration.
        Unit: ms^-2
    """
    g = G*M_plt/(R_plt+h)**2
    return g

@jit(nopython=True)
def calc_hydrostat(P, T, mmw, M_plt, R_plt, H=np.array([])):
    """
    Calculates an altitude profile from given pressure, temperature and
    mean molecular weight profiles assuming hydrostatic equilibrium.


    Parameters
    ----------
    P : ndarray
        Pressure profile
        Unit: Pa
    T : ndarray
        Temperature profile
        Unit: K
    mmw : ndarray
        Mean molecular weight profile
        Unit: kg
    M_plt : real
        Planetary mass
        Unit: kg
    R_plt : real
        Planetary radius
        Unit: m
    H : ndarray
        Altitude profile to be adjusted
        Unit: m

    Returns
    -------
    adjusted_H : ndarray
        Altitude profile satisfying hydrostatic equlibrium.
        Unit: m

    """
    # Note number of profile points and set up a temporary height profile
    NPRO = len(P)
    if len(H)==0:
        H = np.linspace(0,1e6,NPRO)

    # First find level closest ot zero altitude
    ialt = (np.abs(H - 0.0)).argmin()
    alt0 = H[ialt]
    if ( (alt0>0.0) & (ialt>0)):
        ialt = ialt -1

    # iterate until hydrostatic equilibrium
    xdepth = 2
    adjusted_H = H
    dummy_H = np.zeros(NPRO)
    while xdepth > 1:

        h = np.zeros(NPRO)
        dummy_H[:] = adjusted_H

        #Calculating the atmospheric model depth
        atdepth = dummy_H[-1] - dummy_H[0]

        #Calculate the gravity at each altitude level
        gravity = np.zeros(NPRO)
        gravity[:] =  calc_grav_simple(h=dummy_H, M_plt=M_plt, R_plt=R_plt)

        #Calculate the scale height
        scale = np.zeros(NPRO)
        scale[:] = K_B*T[:]/(mmw[:]*gravity[:])

        if ialt > 0 and ialt < NPRO-1 :
            dummy_H[ialt] = 0.0

        # nupper = NPRO - ialt - 1
        for i in range(ialt+1, NPRO):
            sh = 0.5 * (scale[i-1] + scale[i])
            #self.H[i] = self.H[i-1] - sh * np.log(self.P[i]/self.P[i-1])
            dummy_H[i] = dummy_H[i-1] - sh * np.log(P[i]/P[i-1])

        for i in range(ialt-1,-1,-1):
            sh = 0.5 * (scale[i+1] + scale[i])
            #self.H[i] = self.H[i+1] - sh * np.log(self.P[i]/self.P[i+1])
            dummy_H[i] = dummy_H[i+1] - sh * np.log(P[i]/P[i+1])

        atdepth1 = dummy_H[-1] - dummy_H[0]
        xdepth = 100.*abs((atdepth1-atdepth)/atdepth)
        adjusted_H = dummy_H

    return adjusted_H


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Calculate Rayleigh scattering optical path.
"""
from numba import jit
import numpy as np

@jit(nopython=True)
def calc_tau_rayleigh(wave_grid,U_layer,ISPACE=1):
    """
    Calculate the Rayleigh scattering optical depth for Gas Giant atmospheres
    using data from Allen (1976) Astrophysical Quantities.

    Assume H2 ratio of 0.864.

    Parameters
    ----------
    wave_grid(NWAVE) : ndarray
        Wavenumber (cm-1) or wavelength array (um)
    U_layer(NLAYER) : ndarray
        Total absober amount
    ISPACE : int
        Flag indicating the spectral units
        (0) Wavenumber in cm-1 or (1) Wavelegnth (um)

    Outputs
    -------
    tau_rayleigh(NWAVE,NLAYER) : ndarray
        Rayleigh scattering optical path at each wavlength in each layer
    """
    AH2 = 13.58E-5
    BH2 = 7.52E-3
    AHe = 3.48E-5
    BHe = 2.30E-3
    fH2 = 0.864
    k = 1.37971e-23 # JK-1
    P0 = 1.01325e5 # Pa
    T0 = 273.15 # K

    NLAYER = len(U_layer)
    NWAVE = len(wave_grid)

    if ISPACE == 0:
        LAMBDA = 1./wave_grid * 1.0e-2 # converted wavelength unit to m
        x = 1.0/(LAMBDA*1.0e6)
    else:
        LAMBDA = wave_grid * 1.0e-6 # wavelength in m
        x = 1.0/(LAMBDA*1.0e6)

    # calculate refractive index
    nH2 = AH2 * (1.0+BH2*x*x)
    nHe = AHe * (1.0+BHe*x*x)

    #calculate Jupiter air's refractive index at STP (Actually n-1)
    nAir = fH2 * nH2 + (1-fH2)*nHe

    #H2,He Seem pretty isotropic to me?...Hence delta = 0.
    #Penndorf (1957) quotes delta=0.0221 for H2 and 0.025 for He.
    #(From Amundsen's thesis. Amundsen assumes delta=0.02 for H2-He atmospheres
    delta = 0.0
    temp = 32*(np.pi**3.)*nAir**2.
    N0 = P0/(k*T0)

    x = N0*LAMBDA*LAMBDA
    faniso = (6.0+3.0*delta)/(6.0 - 7.0*delta)

    # Calculate the scattering cross sections in m2
    k_rayleighj = temp*faniso/(3.*(x**2)) #(NWAVE)

    # Calculate the Rayleigh opacities in each layer
    tau_rayleigh = np.zeros((len(wave_grid),NLAYER))

    for ilay in range(NLAYER):
        tau_rayleigh[:,ilay] = k_rayleighj[:] * U_layer[ilay] #(NWAVE,NLAYER)

    return tau_rayleigh


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Calculate collision-induced-absorption optical path.
"""
import numpy as np
from numba import jit

@jit(nopython=True)
def calc_tau_cia(wave_grid, K_CIA, ISPACE,
    ID, TOTAM, T_layer, P_layer, VMR_layer, DELH,
    cia_nu_grid, TEMPS, INORMAL, NPAIR=9):
    """
    Calculates
    Parameters
    ----------
    wave_grid : ndarray
        Wavenumber (cm-1) or wavelength array (um) at which to compute
        CIA opacities.
    ID : ndarray
        Gas ID
    # ISO : ndarray
    #     Isotop ID.
    VMR_layer : TYPE
        DESCRIPTION.
    ISPACE : int
        Flag indicating whether the calculation must be performed in
        wavenumbers (0) or wavelength (1)
    K_CIA(NPAIR,NTEMP,NWAVE) : ndarray
         CIA cross sections for each pair at each temperature level and wavenumber.
    cia_nu_grid : TYPE
        DESCRIPTION.
    INORMAL : int


    Returns
    -------
    tau_cia_layer(NWAVE,NLAY) : ndarray
        CIA optical depth in each atmospheric layer.
    """

    # Need to pass NLAY from a atm profile
    NPAIR = 9

    NLAY,NVMR = VMR_layer.shape
    ISO = np.zeros((NVMR))

    # mixing ratios of the relevant gases
    qh2 = np.zeros((NLAY))
    qhe = np.zeros((NLAY))
    qn2 = np.zeros((NLAY))
    qch4 = np.zeros((NLAY))
    qco2 = np.zeros((NLAY))
    # IABSORB = np.ones(5,dtype='int32') * -1

    NWAVEC = 17

    # get mixing ratios from VMR grid
    for iVMR in range(NVMR):
        if ID[iVMR] == 39: # hydrogen
            qh2[:] = VMR_layer[:,iVMR]
            # IABSORB[0] = iVMR
        if ID[iVMR] == 40: # helium
            qhe[:] = VMR_layer[:,iVMR]
            # IABSORB[1] = iVMR
        if ID[iVMR] == 22: # nitrogen
            qn2[:] = VMR_layer[:,iVMR]
            # IABSORB[2] = iVMR
        if ID[iVMR] == 6: # methane
            qch4[:] = VMR_layer[:,iVMR]
            # IABSORB[3] = iVMR
        if ID[iVMR] == 2: # co2
            qco2[:] = VMR_layer[:,iVMR]
            # IABSORB[4] = iVMR

    # calculating the opacity
    XLEN = DELH * 1.0e2 # cm
    TOTAM = TOTAM * 1.0e-4 # cm-2

    ### back to FORTRAN ORIGINAL
    P0=101325
    T0=273.15
    AMAGAT = 2.68675E19 #mol cm-3
    KBOLTZMANN = 1.381E-23
    MODBOLTZA = 10.*KBOLTZMANN/1.013

    tau = (P_layer/P0)**2 * (T0/T_layer)**2 * DELH
    height1 = P_layer * MODBOLTZA * T_layer

    height = XLEN * 1e2
    amag1 = TOTAM /height/AMAGAT
    tau = height*amag1**2

    AMAGAT = 2.68675E19 #mol cm-3
    amag1 = TOTAM / XLEN / AMAGAT # number density
    tau = XLEN*amag1**2# optical path, why fiddle around with XLEN

    # define the calculatiion wavenumbers
    if ISPACE == 0: # input wavegrid is already in wavenumber (cm^-1)
        WAVEN = wave_grid
    elif ISPACE == 1:
        WAVEN = 1.e4/wave_grid
        isort = np.argsort(WAVEN)
        WAVEN = WAVEN[isort] # ascending wavenumbers

    # if WAVEN.min() < cia_nu_grid.min() or WAVEN.max()>cia_nu_grid.max():
    #     print('warning in CIA :: Calculation wavelengths expand a larger range than in .cia file')

    # calculate the CIA opacity at the correct temperature and wavenumber
    NWAVEC = len(wave_grid)  # Number of calculation wavelengths
    tau_cia_layer = np.zeros((NWAVEC,NLAY))

    for ilay in range(NLAY):
        # interpolating to the correct temperature
        temp1 = T_layer[ilay]
        it = (np.abs(TEMPS-temp1)).argmin()

        # want to sandwich the T point
        if TEMPS[it] >= temp1:
            ithi = it
            if it==0:
                # edge case, layer T < T grid
                temp1 = TEMPS[it]
                itl = 0
                ithi = 1
            else:
                itl = it - 1

        elif TEMPS[it]<temp1:
            NT = len(TEMPS)
            itl = it
            if it == NT - 1:
                # edge case, layer T > T grid
                temp1 = TEMPS[it]
                ithi = NT - 1
                itl = NT - 2
            else:
                ithi = it + 1

        # find opacities for the chosen T
        ktlo = K_CIA[:,itl,:]
        kthi = K_CIA[:,ithi,:]

        fhl = (temp1 - TEMPS[itl])/(TEMPS[ithi]-TEMPS[itl])
        fhh = (TEMPS[ithi]-temp1)/(TEMPS[ithi]-TEMPS[itl])

        kt = ktlo * (1.-fhl) + kthi * (1.-fhh)

        # checking that interpolation can be performed to the calculation wavenumbers
        inwave = np.where( (cia_nu_grid>=WAVEN.min()) & (cia_nu_grid<=WAVEN.max()) )
        inwave = inwave[0]

        if len(inwave)>0:

            k_cia = np.zeros((NWAVEC,NPAIR))
            inwave1 = np.where( (WAVEN>=cia_nu_grid.min())&(WAVEN<=cia_nu_grid.max()) )
            inwave1 = inwave1[0]

            for ipair in range(NPAIR):

                # wavenumber interpolation
                # f = interpolate.interp1d(cia_nu_grid,kt[ipair,:])
                # k_cia[inwave1,ipair] = f(WAVEN[inwave1])

                # use numpy for numba integration
                k_cia[inwave1,ipair] = np.interp(WAVEN[inwave1],cia_nu_grid,kt[ipair,:])

            #Combining the CIA absorption of the different pairs (included in .cia file)
            sum1 = np.zeros(NWAVEC)
            if INORMAL==0: # equilibrium hydrogen (1:1)
                sum1[:] = sum1[:] + k_cia[:,0] * qh2[ilay] * qh2[ilay] \
                    + k_cia[:,1] * qhe[ilay] * qh2[ilay]

            elif INORMAL==1: # normal hydrogen (3:1)
                sum1[:] = sum1[:] + k_cia[:,2] * qh2[ilay] * qh2[ilay]\
                    + k_cia[:,3] * qhe[ilay] * qh2[ilay]

            sum1[:] = sum1[:] + k_cia[:,4] * qh2[ilay] * qn2[ilay]
            sum1[:] = sum1[:] + k_cia[:,5] * qn2[ilay] * qch4[ilay]
            sum1[:] = sum1[:] + k_cia[:,6] * qn2[ilay] * qn2[ilay]
            sum1[:] = sum1[:] + k_cia[:,7] * qch4[ilay] * qch4[ilay]
            sum1[:] = sum1[:] + k_cia[:,8] * qh2[ilay] * qch4[ilay]

            # look up CO2-CO2 CIA coefficients (external)
            """
            TO BE DONE
            """
            k_co2 = sum1*0
            # k_co2 = co2cia(WAVEN)

            sum1[:] = sum1[:] + k_co2[:] * qco2[ilay] * qco2[ilay]

            #Look up N2-N2 NIR CIA coefficients
            """
            TO BE DONE
            """
            # TO BE DONE

            #Look up N2-H2 NIR CIA coefficients
            """
            TO BE DONE
            """
            # TO BE DONE

            tau_cia_layer[:,ilay] = sum1[:] * tau[ilay]

    if ISPACE==1:
        tau_cia_layer[:,:] = tau_cia_layer[isort,:]

    return tau_cia_layer


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Reference data files.
"""
# from .calc_mmw import calc_mmw
# from .mol_id import mol_id
# from .mol_info import mol_info
# from .unit import unit, const


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""Routines to read pre-tabulated correlated-k look-up tables (k-tables) and
collision induced absorption opacitied files.

All k-tables are assumed to share the same wavelength grid, pressure grid,
temperature grid, g ordinates and quadrature weights.

Note that the opacities in Nemesis k-tables are multiplied by a factor of 1e20,
which is balanced in the radiative transfer routine by multiplying
absorber amounts by 1e-20.
"""
import numpy as np
from scipy.io import FortranFile
from nemesispy.common.constants import ATM

def read_kta(filepath):
    """
    Reads a pre-tabulated correlated-k look-up table from a Nemesis .kta file.

    Parameters
    ----------
    filepath : str
        The filepath to the Nemesis .kta file to be read.

    Returns
    -------
    gas_id : int
        Gas identifier.
    iso_id : int
        Isotopologue identifier.
    wave_grid(NWAVEKTA) : ndarray
        Wavenumber/wavelength grid of the k-table.
    g_ord(NG) : ndarray
        Quadrature points on the g-ordinates
    del_g(NG) : ndarray
        Gauss quadrature weights for the g-ordinates.
        These are the widths of the bins in g-space.
    P_grid(NPRESSKTA) : ndarray
        Pressure grid on which the k-coefficients are pre-computed.
        Unit: Pa
    T_grid(NTEMPKTA) : ndarray
        Temperature grid on which the k-coefficients are pre-computed.
        Unit: Kelvin
    k_w_g_p_t(NWAVEKTA,NG,NPRESSKTA,NTEMPKTA) : ndarray
        Array storing the k-coefficients.
    """
    # Open file
    if filepath[-3:] == 'kta':
        f = open(filepath,'r')
    else:
        f = open(filepath+'.kta','r')

    # Define bytes consumed by elements of table
    nbytes_int32 = 4
    nbytes_float32 = 4
    ioff = 0 # current position in the file

    # Read headers, irec0 is where ktable data starts
    irec0 = int(np.fromfile(f,dtype='int32',count=1))
    NWAVEKTA = int(np.fromfile(f,dtype='int32',count=1))
    vmin = float(np.fromfile(f,dtype='float32',count=1))
    delv = float(np.fromfile(f,dtype='float32',count=1))
    fwhm = float(np.fromfile(f,dtype='float32',count=1))
    NPRESSKTA = int(np.fromfile(f,dtype='int32',count=1))
    NTEMPKTA = int(np.fromfile(f,dtype='int32',count=1))
    NG = int(np.fromfile(f,dtype='int32',count=1))
    gas_id = int(np.fromfile(f,dtype='int32',count=1))
    iso_id = int(np.fromfile(f,dtype='int32',count=1))

    ioff += 10*nbytes_int32

    # Read g-ordinates and quadrature weights
    g_ord = np.fromfile(f,dtype='float32',count=NG)
    del_g = np.fromfile(f,dtype='float32',count=NG)
    dummy1 = np.fromfile(f,dtype='float32',count=1)
    dummy2 = np.fromfile(f,dtype='float32',count=1)

    ioff += 2*NG*nbytes_float32
    ioff += 2*nbytes_float32

    # Read temperature/pressure grid
    # Note that kta pressure grid is in ATM, conver to Pa
    P_grid = np.fromfile(f,dtype='float32',count=NPRESSKTA) * np.float32(ATM)
    T_grid = np.fromfile(f,dtype='float32',count=NTEMPKTA)
    ioff += NPRESSKTA*nbytes_float32 + NTEMPKTA*nbytes_float32

    # Read wavenumber/wavelength grid
    if delv>0.0:  # uniform grid
        vmax = delv*(NWAVEKTA-1) + vmin
        wave_grid = np.linspace(vmin,vmax,NWAVEKTA)
    else:   # non-uniform grid
        wave_grid = np.zeros((NWAVEKTA))
        wave_grid = np.fromfile(f,dtype='float32',count=NWAVEKTA)
        ioff += NWAVEKTA*nbytes_float32

    # Jump to the minimum wavenumber
    if ioff > irec0-1:
        raise('Error in kta file: too many headers')
    ioff = (irec0-1)*nbytes_float32 # Python index starts at 0
    f.seek(ioff,0)

    # Write the k-coefficients into array form
    k_w_g_p_t = np.zeros((NWAVEKTA,NG,NPRESSKTA,NTEMPKTA))
    k_list = np.fromfile(f,dtype='float32',count=NTEMPKTA*NPRESSKTA*NG*NWAVEKTA)
    ig = 0
    for iwave in range(NWAVEKTA):
        for ipress in range(NPRESSKTA):
            for itemp in range(NTEMPKTA):
                k_w_g_p_t[iwave,:,ipress,itemp] = k_list[ig:ig+NG]
                ig = ig + NG
    k_w_g_p_t = np.float32(k_w_g_p_t)
    # close file
    f.close()
    return gas_id, iso_id, np.float64(wave_grid), g_ord, del_g, P_grid, T_grid, k_w_g_p_t

def read_kls(filepaths):
    """
    Read a list of k-tables from serveral Nemesis .kta files.

    Parameters
    ----------
    filepaths : list
        A list of strings containing filepaths to the .kta files to be read.

    Returns
    -------
    gas_id_list(NGAS) : ndarray
        Gas identifier list.
    iso_id_list(NGAS) : ndarray
        Isotopologue identifier list.
    wave_grid(NWAVEKTA) : ndarray
        Wavenumbers/wavelengths grid of the k-table.
    g_ord(NG) : ndarray
        Quadrature points on the g-ordinates
    del_g(NG) : ndarray
        Gauss quadrature weights for the g-ordinates.
        These are the widths of the bins in g-space.
    P_grid(NPRESSKTA) : ndarray
        Pressure grid on which the k-coefficients are pre-computed.
        Unit: Pa
    T_grid(NTEMPKTA) : ndarray
        Temperature grid on which the k-coefficients are pre-computed.
        Unit: Kelvin
    k_gas_w_g_p_t(NGAS,NWAVEKTA,NG,NPRESSKTA,NTEMPKTA) : ndarray
        Array storing the k-coefficients.

    Notes
    -----
    Assume the k-tables in all the kta files are computed on the same
    wavenumber/wavelength, pressure and temperature grid and with
    same g-ordinates and quadrature weights.
    """
    k_gas_w_g_p_t = []
    gas_id_list = []
    iso_id_list = []
    for filepath in filepaths:
        gas_id, iso_id, wave_grid, g_ord, del_g, P_grid, T_grid,\
          k_g = read_kta(filepath)
        gas_id_list.append(gas_id)
        iso_id_list.append(iso_id)
        k_gas_w_g_p_t.append(k_g)
    # reformat data type to appease numba
    gas_id_list = np.float32(gas_id_list)
    iso_id_list = np.float32(iso_id_list)
    k_gas_w_g_p_t = np.float32(k_gas_w_g_p_t)
    return gas_id_list, iso_id_list, wave_grid, g_ord, del_g, P_grid, T_grid,\
        k_gas_w_g_p_t

def read_cia(filepath,dnu=10,npara=0):
    """
    Parameters
    ----------
    filepath : str
        Filepath to the .tab file containing CIA information.
    dnu : real, optional
        Wavenumber interval. The default is 10.
    npara : int, optional
        DESCRIPTION. The default is 0.

    Returns
    -------
    NU_GRID(NWAVE) : ndarray
        Wavenumber array (NOTE: ALWAYS IN WAVENUMBER, NOT WAVELENGTH).
    TEMPS(NTEMP) : ndarray
        Temperature levels at which the CIA data is defined (K).
    K_CIA(NPAIR,NTEMP,NWAVE) : ndarray
         CIA cross sections for each pair at each temperature level and wavenumber.
    """
    if npara != 0:
        # might need sys.exit'
        raise('Routines have not been adapted yet for npara!=0')

    # Reading the actual CIA file
    if npara == 0:
        NPAIR = 9 # 9 pairs of collision induced absorption opacities

    f = FortranFile(filepath, 'r')
    TEMPS = f.read_reals( dtype='float64' )
    KCIA_list = f.read_reals( dtype='float32' )
    NT = len(TEMPS)
    NWAVE = int(len(KCIA_list)/NT/NPAIR)
    NU_GRID = np.linspace(0,dnu*(NWAVE-1),NWAVE)
    K_CIA = np.zeros((NPAIR,NT,NWAVE)) # NPAIR x NT x NWAVE

    index = 0
    for iwn in range(NWAVE):
        for itemp in range(NT):
            for ipair in range(NPAIR):
                K_CIA[ipair,itemp,iwn] = KCIA_list[index]
                index += 1

    return NU_GRID, TEMPS, K_CIA


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Constants used in this package.
Follow the CODATA Recommended Values of Fundamental Physical Constants.
J. Phys. Chem. Ref. Data 50, 033105 (2021);
https://doi.org/10.1063/5.0064853

from nemesispy.common.constants import R_SUN,M_SUN
"""
# CODATA
C_LIGHT = 299792458       # ms-1 speed of light in vacuum
G = 6.67430e-11           # m3 kg-1 s-2 Newtonian constant of gravitation
PLANCK = 6.62607015e-34   # Js Planck constant
K_B = 1.380649e-23        # J K-1 Boltzmann constant
N_A = 6.02214076e23       # Avagadro's number
SIGMA_SB = 5.670374419e-8 # Wm-2K-4 Stefan-Boltzmann constant
AMU = 1.66053906660e-27   # kg atomic mass unit

# Astronomical units
AU = 1.495978707e+11      # m astronomical unit
R_SUN = 6.95700e8         # m solar radius
R_JUP = 6.9911e7          # m Jupiter radius
R_JUP_E = 7.1492e7        # m nominal equatorial Jupiter radius
M_SUN = 1.98847542e+30    # kg solar mass
M_JUP = 1.898e27          # kg Jupiter mass

# Others
ATM = 101325              # Pa atmospheric pressure

# Derived
C1 = 1.1910429723971884e-12 # W cm2 2*PLANCK*C_LIGHT**2
C2 = 1.4387768775039337 # cm K-1 PLANCK*C_LIGHT/K_B


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Routines to split an atmosphere into layers and calculate the
average layer properties along an observing path.
"""
import numpy as np
from scipy.integrate import simps
# from numba import jit
from nemesispy.common.constants import K_B
from nemesispy.radtran.calc_mmw import calc_mmw

# @jit(nopython=True)
def split(H_model, P_model, NLAYER, layer_type=1, H_0=0.0,
    planet_radius=None, custom_path_angle=0.0,
    custom_H_base=np.array([0,0]), custom_P_base=np.array([0,0])):
    """
    Splits atmospheric models into layers. Returns layer base altitudes
    and layer base pressures.

    Parameters
    ----------
    H_model(NMODEL) : ndarray
        Altitudes at which the atmospheric models are defined.
        Unit: m
    P_model(NMODEL) : ndarray
        Pressures at which the atmospheric models arer defined.
        (At altitude H_model[i] the pressure is  P_model[i].)
        Unit: Pa
    NLAYER : int
        Number of layers to split the atmospheric models into.
    layer_type : int, optional
        Integer specifying how to split up the layers.
        0 = split by equal changes in pressure
        1 = split by equal changes in log pressure
        2 = split by equal changes in height
        3 = split by equal changes in path length
        4 = split by layer base pressure levels specified in P_base
        5 = split by layer base height levels specified in H_base
        Note 4 and 5 force NLAYER = len(P_base) or len(H_base).
        The default is 1.
    H_0 : real, optional
        Altitude of the lowest point in the atmospheric model.
        This is defined with respect to the reference planetary radius, i.e.
        the altitude at planet_radius is 0.
        The default is 0.0.
    planet_radius : real, optional
        Reference planetary planet_radius where H_model is set to be 0.  Usually
        set at surface for terrestrial planets, or at 1 bar pressure level for
        gas giants.
        Required only for layer type 3.
        The default is None.
    custom_path_angle : real, optional
        Required only for layer type 3.
        Zenith angle in degrees defined at the base of the lowest layer.
        The default is 0.0.
    custom_H_base(NLAYER) : ndarray, optional
        Required only for layer type 5.
        Altitudes of the layer bases defined by user.
        The default is None.
    custom_P_base(NLAYER) : ndarray, optional
        Required only for layer type 4.
        Pressures of the layer bases defined by user.
        The default is None.

    Returns
    -------
    H_base(NLAYER) : ndarray
        Heights of the layer bases.
    P_base(NLAYER) : ndarray
        Pressures of the layer bases.
    """
    assert (H_0>=H_model[0]) and (H_0<H_model[-1]) , \
        'Lowest layer base altitude not contained in atmospheric model'
    if layer_type == 0: # split by equal pressure intervals
        # interpolate for the pressure at base of lowest layer
        bottom_pressure = np.interp(H_0, H_model,P_model)
        P_base = np.linspace(bottom_pressure,P_model[-1],NLAYER+1)[:-1]
        # np.interp need x-coor to be increasing
        P_model = P_model[::-1]
        H_model = H_model[::-1]
        H_base = np.interp(P_base,P_model,H_model)

    elif layer_type == 1: # split by equal log pressure intervals
        # interpolate for the pressure at base of lowest layer
        bottom_pressure = np.interp(H_0,H_model,P_model)
        # P_base = np.logspace(np.log10(bottom_pressure),np.log10(P_model[-1]),\
        #     NLAYER+1)[:-1]

        P_base = 10**np.linspace(np.log10(bottom_pressure),np.log10(P_model[-1]),\
            NLAYER+1)[:-1]

        # np.interp need ascending x-coord
        P_model = P_model[::-1]
        H_model = H_model[::-1]
        H_base = np.interp(P_base,P_model,H_model)

    elif layer_type == 2: # split by equal height intervals
        H_base = np.linspace(H_model[0]+H_0, H_model[-1], NLAYER+1)[:-1]
        P_base = np.interp(H_base,H_model,P_model)

    elif layer_type == 3: # split by equal line-of-sight path intervals
        assert custom_path_angle<=90 and custom_path_angle>=0,\
            'Zennith angle should be in range [0,90] degree'
        sin = np.sin(custom_path_angle*np.pi/180) # sin(custom_path_angle angle)
        cos = np.cos(custom_path_angle*np.pi/180) # cos(custom_path_angle angle)
        r0 = planet_radius+H_0 # radial distance to lowest layer's base
        rmax = planet_radius+H_model[-1] # radial distance maximum height
        S_max = np.sqrt(rmax**2-(r0*sin)**2)-r0*cos # total path length
        S_base = np.linspace(0, S_max, NLAYER+1)[:-1]
        H_base = np.sqrt(S_base**2+r0**2+2*S_base*r0*cos)-planet_radius
        logP_base = np.interp(H_base,H_model,np.log(P_model))
        P_base = np.exp(logP_base)

    elif layer_type == 4: # split by specifying input base pressures
        # assert np.all(custom_P_base!=None),'Need input layer base pressures'
        assert  (custom_P_base[-1] > P_model[-1]) \
            and (custom_P_base[0] <= P_model[0]), \
            'Input layer base pressures out of range of atmosphere profile'
        NLAYER = len(custom_P_base)
        # np.interp need ascending x-coord
        P_model = P_model[::-1]
        H_model = H_model[::-1]
        H_base = np.interp(custom_P_base,P_model,H_model)

    elif layer_type == 5: # split by specifying input base heights
        # assert np.all(custom_H_base!=None), 'Need input layer base heighs'
        assert (custom_H_base[-1] < H_model[-1]) \
            and (custom_H_base[0] >= H_model[0]), \
            'Input layer base heights out of range of atmosphere profile'
        NLAYER = len(custom_H_base)
        P_base = np.interp(custom_H_base,H_model,P_model)
    else:
        raise Exception('Layering scheme not defined')
    return H_base, P_base

# @jit(nopython=True)
def average(planet_radius, H_model, P_model, T_model, VMR_model, ID, H_base,
        path_angle, H_0=0.0, NSIMPS=101):
    """
    Calculates absorber-amount-weighted average layer properties of an
    atmosphere.

    Inputs
    ------
    planet_radius : real
        Reference planetary planet_radius where H_model is set to be 0.  Usually
        set at surface for terrestrial planets, or at 1 bar pressure level for
        gas giants.
    H_model(NMODEL) : ndarray
        Altitudes at which the atmospheric model is defined.
        (At altitude H_model[i] the pressure is P_model[i].)
        Unit: m
    P_model(NMODEL) : ndarray
        Pressures at which the atmospheric model is defined.
        (At altitude P_model[i] the pressure is H_model[i].)
        Unit: Pa
    T_mode(NMODEL) : ndarray
        Temperature profile defined in the atmospheric model.
    ID : ndarray
        Gas identifiers.
    VMR_model(NMODEL,NGAS) : ndarray
        Volume mixing ratios of gases defined in the atmospheric model.
        VMR_model[i,j] is Volume Mixing Ratio of jth gas at ith profile point.
        The jth column corresponds to the gas with RADTRANS ID ID[j].
    H_base(NLAYER) : ndarray
        Heights of the layer bases.
    path_angle : real
        Zenith angle in degrees defined at H_0.
    H_0 : real, default 0.0
        Altitude of the lowest point in the atmospheric model.
        This is defined with respect to the reference planetary radius, i.e.
        the altitude at planet_radius is 0.
    NSIMPS : int, optional
        Number of Simpson's integration points to be used if integration_type=1.
        The default is 101.

    Returns
    -------
    H_layer(NLAYER) : ndarray
        Representative height for each layer
        Unit: m
    P_layer(NLAYER) : ndarray
        Representative pressure for each layer
        Unit: Pa
    T_layer(NLAYER) : ndarray
        Representative pressure for each layer
        Unit: Kelven
    U_layer(NLAYER) : ndarray
        Total gaseous absorber amounts along the line-of-sight path, i.e.
        total number of gas molecules per unit area.
        Unit: no of absorber per m^2
    VMR_layer(NLAYER, NGAS) : ndarray
        Representative partial pressure for each gas at each layer.
        VMR_layer[i,j] is representative partial pressure of gas j in layer i.
    Gas_layer(NLAYER, NGAS) : ndarray
        Representative absorber amounts of each gas at each layer.
        Gas_layer[i,j] is the representative number of gas j molecules
        in layer i in the form of number of molecules per unit area.
        Unit: no of absorber per m^2
    scale(NLAYER) : ndarray
        Layer scaling factor, i.e. ratio of path length through each layer
        to the layer thickness.
    del_S(NLAYER) : ndarray
        Path lengths.
        Unit: m

    Notes
    -----
    Assume SI units.
    """
    # Calculate layer geometric properties
    NLAYER = len(H_base)
    del_H = np.concatenate(((H_base[1:]-H_base[:-1]),[H_model[-1]-H_base[-1]]))
    sin = np.sin(path_angle*np.pi/180) # sin(viewing angle)
    cos = np.cos(path_angle*np.pi/180) # cos(viewing angle)
    r0 = planet_radius+H_0 # minimum radial distance
    rmax = planet_radius+H_model[-1] # maximum radial distance
    S_max = np.sqrt(rmax**2-(r0*sin)**2)-r0*cos # total path length
    S_base = np.sqrt((planet_radius+H_base)**2-(r0*sin)**2)-r0*cos # path lengths at base of layer
    del_S = np.concatenate(((S_base[1:]-S_base[:-1]),[S_max-S_base[-1]]))
    scale = del_S/del_H # Layer Scaling Factor

    # initiate output arrays
    Ngas = len(VMR_model[0])
    H_layer = np.zeros(NLAYER) # average layer height
    P_layer = np.zeros(NLAYER) # average layer pressure
    T_layer = np.zeros(NLAYER) # average layer temperature
    U_layer = np.zeros(NLAYER) # total no. of gas molecules per unit aera
    dU_dS = np.zeros(NLAYER) # no. of gas molecules per area per distance
    Gas_layer = np.zeros((NLAYER, Ngas)) # no. of molecules per aera
    VMR_layer = np.zeros((NLAYER, Ngas)) # partial pressures
    MMW_layer = np.zeros(NLAYER) # mean molecular weight

    # Calculate average properties depending on intergration type
    # use absorber-amount-weighted averages calculated with Simpsons rule
    for ilayer in range(NLAYER):
        S0 = S_base[ilayer]
        if ilayer < NLAYER-1:
            S1 = S_base[ilayer+1]
        else:
            S1 = S_max
        # sub-divide each layer into NSIMPS layers for integration
        S_int = np.linspace(S0, S1, NSIMPS)
        H_int = np.sqrt(S_int**2+r0**2+2*S_int*r0*cos)-planet_radius
        P_int = np.interp(H_int,H_model,P_model)
        T_int = np.interp(H_int,H_model,T_model)
        dU_dS_int = P_int/(K_B*T_int)
        VMR_int = np.zeros((NSIMPS, Ngas))
        MMW_int = np.zeros(NSIMPS)

        # absorber amount weighted integrals
        U_layer[ilayer] = simps(dU_dS_int,S_int)
        H_layer[ilayer] = simps(H_int*dU_dS_int,S_int)/U_layer[ilayer]
        P_layer[ilayer] = simps(P_int*dU_dS_int,S_int)/U_layer[ilayer]
        T_layer[ilayer] = simps(T_int*dU_dS_int,S_int)/U_layer[ilayer]
        for J in range(Ngas):
            VMR_int[:,J] = np.interp(H_int,H_model,VMR_model[:,J])
            Gas_layer[ilayer,J] = simps(VMR_int[:,J]*dU_dS_int,S_int)
        # VMR_int = (Gas_int.T * P_int).T # gas partial pressures
        for J in range(Ngas):
            VMR_layer[ilayer, J] \
                = simps(VMR_int[:,J]*dU_dS_int,S_int)/U_layer[ilayer]
        for K in range(NSIMPS):
            MMW_int[K] = calc_mmw(ID, VMR_int[K,:])
        MMW_layer[ilayer] = simps(MMW_int*dU_dS_int,S_int)/U_layer[ilayer]

    # Scale back to vertical layers
    U_layer = U_layer / scale
    Gas_layer = (Gas_layer.T * scale**-1 ).T

    return H_layer,P_layer,T_layer,VMR_layer,U_layer,Gas_layer,MMW_layer,\
        scale,del_S,del_H

# @jit(nopython=True)
def calc_layer(planet_radius, H_model, P_model, T_model, VMR_model, ID, NLAYER,
    path_angle, H_0=0.0, NSIMPS=101,
    layer_type=1, custom_path_angle=0.0,
    custom_H_base=None, custom_P_base=None):
    """
    Top level routine that calculates the layer properties from an atmospehric
    model.

    Parameters
    ----------
    planet_radius : real
        Reference planetary planet_radius where H_model is set to be 0.  Usually
        set at surface for terrestrial planets, or at 1 bar pressure level for
        gas giants.
    H_model(NMODEL) : ndarray
        Altitudes at which the atmospheric model is defined.
        (At altitude H_model[i] the pressure is P_model[i].)
        Unit: m
    P_model(NMODEL) : ndarray
        Pressures at which the atmospheric model is defined.
        (At altitude P_model[i] the pressure is H_model[i].)
        Unit: Pa
    T_mode(NMODEL) : ndarray
        Temperature profile defined in the atmospheric model.
    VMR_model : ndarray
        Volume mixing ratios of gases defined in the atmospheric model.
        VMR_model[i,j] is the Volume Mixing Ratio of jth gas at ith profile point.
        The jth column corresponds to the gas with RADTRANS ID ID[j].
    ID : ndarray
        Gas identifiers.
    NLAYER : int
        Number of layers to split the atmospheric model into.
    path_angle : real
        Zenith angle in degrees defined at H_0.
    H_0 : real, default 0.0
        Altitude of the lowest point in the atmospheric model.
        This is defined with respect to the reference planetary radius, i.e.
        the altitude at planet_radius is 0.
    layer_type : int, default 1
        Integer specifying how to split up the layers.
        0 = split by equal changes in pressure
        1 = split by equal changes in log pressure
        2 = split by equal changes in height
        3 = split by equal changes in path length
        4 = split by layer base pressure levels specified in P_base
        5 = split by layer base height levels specified in H_base
        Note 4 and 5 force NLAYER = len(P_base) or len(H_base).
    NSIMPS : int, optional
        Number of Simpson's integration points to be used if integration_type=1.
        The default is 101.
    custom_path_angle : real, optional
        Required only for layer type 3.
        Zenith angle in degrees defined at the base of the lowest layer.
        The default is 0.0.
    custom_H_base(NLAYER) : ndarray, optional
        Required only for layer type 5.
        Altitudes of the layer bases defined by user.
        The default is None.
    custom_P_base(NLAYER) : ndarray, optional
        Required only for layer type 4.
        Pressures of the layer bases defined by user.
        The default is None.


    Returns
    -------
    H_layer(NLAYER) : ndarray
        Representative height for each layer
        Unit: m
    P_layer(NLAYER) : ndarray
        Representative pressure for each layer
        Unit: Pa
    T_layer(NLAYER) : ndarray
        Representative pressure for each layer
        Unit: Kelven
    U_layer(NLAYER) : ndarray
        Total gaseous absorber amounts along the line-of-sight path, i.e.
        total number of gas molecules per unit area.
        Unit: no of absorber per m^2
    VMR_layer(NLAYER, NGAS) : ndarray
        Representative partial pressure for each gas at each layer.
        VMR_layer[i,j] is representative partial pressure of gas j in layer i.
    Gas_layer(NLAYER, NGAS) : ndarray
        Representative absorber amounts of each gas at each layer.
        Gas_layer[i,j] is the representative number of gas j molecules
        in layer i in the form of number of molecules per unit area.
        Unit: no of absorber per m^2
    scale(NLAYER) : ndarray
        Layer scaling factor, i.e. ratio of path length through each layer
        to the layer thickness.
    del_S(NLAYER) : ndarray
        Path lengths.
        Unit: m
    """
    H_base, P_base = split(H_model, P_model, NLAYER, layer_type=layer_type,
        H_0=H_0, planet_radius=planet_radius,
        custom_path_angle=custom_path_angle, custom_H_base=custom_H_base,
        custom_P_base=custom_P_base)

    H_layer,P_layer,T_layer,VMR_layer,U_layer,Gas_layer,MMW_layer,\
        scale,del_S,del_H = average(planet_radius, H_model, P_model, T_model,
            VMR_model, ID, H_base, path_angle=path_angle,
            H_0=H_0, NSIMPS=NSIMPS)

    return H_layer,P_layer,T_layer,VMR_layer,U_layer,Gas_layer,MMW_layer,\
        scale,del_S,del_H


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from numba import jit

@jit(nopython=True)
def VERINT(X,Y,XIN):
    """
    np.interp only works with strictly increasing independent variable.
    """
    if X[0]>X[-1]:
        X = X[::-1]
        Y = Y[::-1]
    YOUT = np.interp(x=XIN,xp=X,fp=Y)
    return YOUT

@jit(nopython=True)
def interpvivien_point(lon, lat, p, gcm_lon, gcm_lat, gcm_p, gcm_t, gcm_vmr,
    substellar_point_longitude_shift=0):
    """
    Find the T(P) profile and VMR(P) profile at a location specified by (lon,lat)
    by interpolating a GCM.

    Parameters
    ----------
    lon : real
        Longitude of the location
    lat : real
        Latitude of the location
    p : ndarray
        Pressure grid for the output T(P) and VMR(P) profile
    gcm_lon : ndarray
        Longitude grid of the GCM, assumed to be [-180,180] and increasing. (1D)
    gcm_lat : ndarray
        Latitude grid of the GCM, assumed to be [-90,90] and increasing. (1D)
    gcm_p : ndarray
       Pressure grid of the GCM. (1D)
    gcm_t : ndarray
        Temperature model.
        Has dimension NLON x NLAT x NPRESS.
    gcm_vmr : ndarray
        VMR model.
        NLON x NLAT x NPRESS x NVMR
    substellar_point_longitude_shift : real
        The longitude shift between the longitude-lattitude coordinate system
        of the GCM model to the output coordinate system. For example, if in the
        output coordinate system the substellar point is defined at 0 E,
        whereas in the GCM coordinate system the substellar point is defined
        at 90 E, put substellar_point_longitude_shift=90.

    Returns
    -------
    T : ndarray
        Temperature model interpolated to (lon,lat,p)
    VMR : ndarray
        VMR model interpolated to (lon,lat,p)

    """
    # Check input latitude make sense
    assert lat<=90 and lat>=-90

    # Number of pressures in the interped profile
    NPRO = len(p)

    # Dimension parameters of the atmospheric model
    NLON,NLAT,NPRESS,NVMR = gcm_vmr.shape

    # Convert to log pressure
    log_gcm_P = np.log(gcm_p)

    # Shift to desired longitude coordinate system
    # while preserving the monotonicity of longitude
    if substellar_point_longitude_shift != 0:
        substellar_point_longitude_shift \
            = np.mod(substellar_point_longitude_shift,360)
        if substellar_point_longitude_shift <=180:
            gcm_lon = gcm_lon + substellar_point_longitude_shift
        else:
            gcm_lon = gcm_lon - (360-substellar_point_longitude_shift)

    #  Find closest point in stored array
    JLAT = -1
    for I in range(NLAT-1):
        if lat >= gcm_lat[I] and lat <= gcm_lat[I+1]:
            JLAT = I
            FLAT = (lat-gcm_lat[I])/(gcm_lat[I+1]-gcm_lat[I])
    if JLAT < 0:
        if lat < gcm_lat[0]:
            JLAT = 0
            FLAT = 0
        if lat >= gcm_lat[-1]:
            JLAT = NLAT - 1
            FLAT = 1

    JLON1 = -1
    JLON2 = -1
    for I in range(NLON-1):
        if lon >= gcm_lon[I] and lon <= gcm_lon[I+1]:
            JLON1 = I
            JLON2 = I+1
            FLON = (lon-gcm_lon[I])/(gcm_lon[I+1]-gcm_lon[I])
    if JLON1 < 0:
        if lon < gcm_lon[0]:
            # lon must be in range 0. to gcm_lon[0]
            JLON1 = NLON - 1
            JLON2 = 0
            FLON = (lon+360-gcm_lon[-1])/(gcm_lon[0]+360-gcm_lon[-1])
        if lon >= gcm_lon[-1]:
            # lon must be in range gcm_lon[-1] to 360
            JLON1 = NLON - 1
            JLON2 = 0
            FLON = (lon - gcm_lon[-1])/(gcm_lon[0]+360-gcm_lon[-1])

    # Output arrays
    interped_T = np.zeros(NPRO)
    interped_VMR = np.zeros((NPRO,NVMR))

    # Temperature interpolation array
    tempVY1 = gcm_t[JLON1,JLAT,:]
    tempVY2 = gcm_t[JLON2,JLAT,:]
    tempVY3 = gcm_t[JLON2,JLAT+1,:]
    tempVY4 = gcm_t[JLON1,JLAT+1,:]
    for IPRO in range(NPRO):
        # convert pressure to atm then to log
        LP1 = np.log(p[IPRO])
        tempY1 = VERINT(log_gcm_P,tempVY1,LP1)
        tempY2 = VERINT(log_gcm_P,tempVY2,LP1)
        tempY3 = VERINT(log_gcm_P,tempVY3,LP1)
        tempY4 = VERINT(log_gcm_P,tempVY4,LP1)
        interped_T[IPRO] = (1.0-FLON)*(1.0-FLAT)*tempY1 + FLON*(1.0-FLAT)*tempY2 \
            + FLON*FLAT*tempY3 + (1.0-FLON)*FLAT*tempY4
        for IVMR in range(NVMR):
            gasVY1 = gcm_vmr[JLON1,JLAT,:,IVMR]
            gasVY2 = gcm_vmr[JLON2,JLAT,:,IVMR]
            gasVY3 = gcm_vmr[JLON2,JLAT+1,:,IVMR]
            gasVY4 = gcm_vmr[JLON1,JLAT+1,:,IVMR]
            gasY1 = VERINT(log_gcm_P,gasVY1,LP1)
            gasY2 = VERINT(log_gcm_P,gasVY2,LP1)
            gasY3 = VERINT(log_gcm_P,gasVY3,LP1)
            gasY4 = VERINT(log_gcm_P,gasVY4,LP1)
            interped_VMR[IPRO,IVMR] = (1.0-FLON)*(1.0-FLAT)*gasY1 + FLON*(1.0-FLAT)*gasY2\
                + FLON*FLAT*gasY3 + (1.0-FLON)*FLAT*gasY4

    return interped_T, interped_VMR

def interp_gcm_X(lon, lat, p, gcm_lon, gcm_lat, gcm_p, X,
    substellar_point_longitude_shift=0):
    """
    Find the X(P) profile at a location specified by (lon,lat) by interpolating
    a GCM. X can be any scalar quantity modeled in the GCM.

    Parameters
    ----------
    lon : real
        Longitude of the location
    lat : real
        Latitude of the location
    p : ndarray
        Pressure grid for the output X(P) profile
    gcm_lon : ndarray
        Longitude grid of the GCM, assumed to be [-180,180] and increasing. (1D)
    gcm_lat : ndarray
        Latitude grid of the GCM, assumed to be [-90,90] and increasing. (1D)
    gcm_p : ndarrray
        Pressure grid of the GCM. (1D)
    X : ndarray
        A scalar quantity defined in the GCM, e.g., temperature or VMR. (3D)
        Has dimensioin NLON x NLAT x NP
    substellar_point_longitude_shift : real
        The longitude shift between the longitude-lattitude coordinate system
        of the GCM model to the output coordinate system. For example, if in the
        output coordinate system the substellar point is defined at 0 E,
        whereas in the GCM coordinate system the substellar point is defined
        at 90 E, put substellar_point_longitude_shift=90.

    Returns
    -------
    interped_X : ndarray
        X interpolated to (lon,lat,p).
    """
    # Check input latitude make sense
    assert lat <= 90 and lat >= -90

    # Number of pressures in the interped profile
    NPRO = len(p)

    # Dimensions of GCM
    NLON = len(gcm_lon)
    NLAT = len(gcm_lat)
    NPRESS = len(gcm_p)

    # Work in log pressure
    log_gcm_p = np.log(gcm_p)

    # Shift to desired longitude coordinate system
    # while preserving the monotonicity of longitude
    if substellar_point_longitude_shift != 0:
        substellar_point_longitude_shift \
            = np.mod(substellar_point_longitude_shift,360)
        if substellar_point_longitude_shift <=180:
            gcm_lon = gcm_lon + substellar_point_longitude_shift
        else:
            gcm_lon = gcm_lon - (360-substellar_point_longitude_shift)

    #  Find closest point on the GCM grid to input location
    JLAT = -1
    for I in range(NLAT-1):
        if lat >= gcm_lat[I] and lat <= gcm_lat[I+1]:
            JLAT = I
            FLAT = (lat-gcm_lat[I])/(gcm_lat[I+1]-gcm_lat[I])
    if JLAT < 0:
        if lat < gcm_lat[0]:
            JLAT = 0
            FLAT = 0
        if lat >= gcm_lat[-1]:
            JLAT = NLAT - 1
            FLAT = 1

    JLON1 = -1
    JLON2 = -1
    for I in range(NLON-1):
        if lon >= gcm_lon[I] and lon <= gcm_lon[I+1]:
            JLON1 = I
            JLON2 = I+1
            FLON = (lon-gcm_lon[I])/(gcm_lon[I+1]-gcm_lon[I])
    if JLON1 < 0:
        if lon < gcm_lon[0]:
            # lon must be in range 0. to gcm_lon[0]
            JLON1 = NLON - 1
            JLON2 = 0
            FLON = (lon+360-gcm_lon[-1])/(gcm_lon[0]+360-gcm_lon[-1])
        if lon >= gcm_lon[-1]:
            # lon must be in range gcm_lon[-1] to 360
            JLON1 = NLON - 1
            JLON2 = 0
            FLON = (lon - gcm_lon[-1])/(gcm_lon[0]+360-gcm_lon[-1])

    # Output arrays
    interped_X = np.zeros(NPRO)

    # Interpolation array
    VY1 = X[JLON1,JLAT,:]
    VY2 = X[JLON2,JLAT,:]
    VY3 = X[JLON2,JLAT+1,:]
    VY4 = X[JLON1,JLAT+1,:]
    log_p = np.log(p)
    for IPRO in range(NPRO):
        # convert pressure to atm then to log
        LP1 = log_p[IPRO]
        Y1 = VERINT(log_gcm_p,VY1,LP1)
        Y2 = VERINT(log_gcm_p,VY2,LP1)
        Y3 = VERINT(log_gcm_p,VY3,LP1)
        Y4 = VERINT(log_gcm_p,VY4,LP1)
        interped_X[IPRO] = (1.0-FLON)*(1.0-FLAT)*Y1 + FLON*(1.0-FLAT)*Y2 \
            + FLON*FLAT*Y3 + (1.0-FLON)*FLAT*Y4

    return interped_X



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import os
import time
from nemesispy.radtran.forward_model import ForwardModel
matplotlib.interactive(True)
# Read GCM data
from nemesispy.data.gcm.process_gcm import (nlon,nlat,xlon,xlat,npv,pv,\
    tmap,h2omap,comap,co2map,ch4map,hemap,h2map,vmrmap,\
    tmap_mod,h2omap_mod,comap_mod,co2map_mod,ch4map_mod,\
    hemap_mod,h2map_mod,vmrmap_mod,phase_grid,\
    kevin_phase_by_wave,kevin_wave_by_phase,\
    pat_phase_by_wave,pat_wave_by_phase,\
    vmrmap_mod_new,tmap_hot)

from nemesispy.common.helper import lowres_file_paths, cia_file_path

print('creating example phase curve')
### Wavelengths grid and orbital phase grid
wave_grid = np.array([1.1425, 1.1775, 1.2125, 1.2475, 1.2825, 1.3175, 1.3525, 1.3875,
       1.4225, 1.4575, 1.4925, 1.5275, 1.5625, 1.5975, 1.6325, 3.6   ,
       4.5   ])
phase_grid = np.array([ 22.5,  45. ,  67.5,  90. , 112.5, 135. , 157.5, 180. , 202.5,
       225. , 247.5, 270. , 292.5, 315. , 337.5])
nwave = len(wave_grid)
nphase = len(phase_grid)
wasp43_spec = np.array([3.341320e+25, 3.215455e+25, 3.101460e+25, 2.987110e+25,
       2.843440e+25, 2.738320e+25, 2.679875e+25, 2.598525e+25,
       2.505735e+25, 2.452230e+25, 2.391140e+25, 2.345905e+25,
       2.283720e+25, 2.203690e+25, 2.136015e+25, 1.234010e+24,
       4.422200e+23])

### Reference Planet Input
M_plt = 3.8951064000000004e+27 # kg
R_plt = 74065.70 * 1e3 # m
gas_id = np.array([  1, 2,  5,  6, 40, 39])
iso_id = np.array([0, 0, 0, 0, 0, 0])
NLAYER = 20
phasenumber = 3
nmu = 3
phase = phase_grid[phasenumber]
P_model = np.geomspace(20e5,100,NLAYER)
NITER = 1
### Set up forward model
FM = ForwardModel()
FM.set_planet_model(M_plt=M_plt,R_plt=R_plt,gas_id_list=gas_id,iso_id_list=iso_id,
    NLAYER=NLAYER)
FM.set_opacity_data(kta_file_paths=lowres_file_paths, cia_file_path=cia_file_path)

### Testing one particular orbital phase (inhomogeneouus disc averaging)
start1 = time.time()
one_phase =  FM.calc_disc_spectrum(phase=phase, nmu=nmu, P_model = P_model,
    global_model_P_grid=pv,
    global_T_model=tmap_mod, global_VMR_model=vmrmap_mod,
    mod_lon=xlon,
    mod_lat=xlat,
    solspec=wasp43_spec)
end1 = time.time()
print('compile+run time = ',end1-start1)

start2 = time.time()
one_phase =  FM.calc_disc_spectrum(phase=phase, nmu=nmu, P_model = P_model,
    global_model_P_grid=pv,
    global_T_model=tmap_mod, global_VMR_model=vmrmap_mod,
    mod_lon=xlon,
    mod_lat=xlat,
    solspec=wasp43_spec)
end2 = time.time()
print('run time = ',end2-start2)

fig, axs = plt.subplots(nrows=2,ncols=1,sharex=True,
    dpi=100)
axs[0].set_title('phase = {}'.format(phase))
axs[0].plot(wave_grid,one_phase,color='b',
    linewidth=0.5,linestyle='--',
    marker='x',markersize=2,label='Python')
axs[0].scatter(wave_grid,kevin_phase_by_wave[phasenumber,:,0],color='r',marker='+',label='Data')
axs[0].plot(wave_grid,pat_phase_by_wave[phasenumber],color='k',
    linewidth=0.5,linestyle='-',
    marker='x',markersize=2,label='Fortran')
axs[0].legend()
axs[0].grid()

diff = (one_phase - pat_phase_by_wave[phasenumber,:])/one_phase
axs[1].scatter(wave_grid,diff,marker='.',color='b')
axs[1].grid()
print(diff)

plt.savefig('create_example.pdf')

"""
### This is for plotting specta at all phases
for iphase in range(nphase):
    phasenumber = iphase
    nmu = 5
    phase = phase_grid[phasenumber]
    P_model = np.geomspace(20e5,1,NLAYER)
    P_model = pv
    one_phase =  FM.test_disc_spectrum(phase=phase, nmu=nmu, P_model = pv,
        global_model_P_grid=pv,
        global_T_model=tmap, global_VMR_model=vmrmap_mod_new,
        model_longitudes=xlon,
        model_lattitudes=xlat,
        solspec=wasp43_spec)

    fig, axs = plt.subplots(nrows=2,ncols=1,sharex=True,
        dpi=800)
    axs[0].set_title('phase = {}'.format(phase))
    axs[0].plot(wave_grid,one_phase,color='b',label='Python')
    axs[0].scatter(wave_grid,kevin_phase_by_wave[phasenumber,:,0],color='r',marker='+',label='Data')
    axs[0].plot(wave_grid,pat_phase_by_wave[phasenumber,:],color ='k',label='Fortran')
    axs[0].legend(loc='upper left')
    axs[0].grid()
    axs[0].set_ylabel('Flux ratio')

    diff = (one_phase - pat_phase_by_wave[phasenumber,:])/one_phase
    axs[1].scatter(wave_grid,diff,marker='.',color='b')
    axs[1].grid()
    axs[1].set_ylabel('Relative diff')
    axs[1].set_xlabel('Wavelength (Micron)')
    print(iphase,diff)
    plt.tight_layout()

    plt.show()
    # plt.savefig('good_discav_planet{}.pdf'.format(iphase),dpi=800)

"""



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Calculate the Planck function.
Note that in the Fortran Nemesis code, the values of constants used are
C1 = 1.1911e-12 W cm2
C2 = 1.439 cm K-1
"""
import numpy as np
from numba import jit

@jit(nopython=True)
def calc_planck(wave_grid,T,ispace=1):
    """
    Calculates the blackbody radiation.

    Parameters
    ----------
    wave_grid(nwave) : ndarray
        Wavelength or wavenumber array
        Unit: um or cm-1
    T : real
        Temperature of the blackbody (K)
    ispace : int
        Flag indicating the spectral units
        (0) Wavenumber (cm-1)
        (1) Wavelength (um)

    Returns
    -------
	bb(nwave) : ndarray
        Spectral radiance.
        Unit: (0) W cm-2 sr-1 (cm-1)-1
              (1) W cm-2 sr-1 um-1
    """
    wave = wave_grid.astype(np.float64)
    if np.any(wave<=0) or T < 0:
        raise(Exception('error in calc_planck: negative wavelengths' \
            +' or temperature'))
    C1 = np.array([1.1910e-12]) # W cm2 2*PLANCK*C_LIGHT**2
    C2 = np.array([1.4388]) # cm K-1 PLANCK*C_LIGHT/K_B
    if ispace==0:
        y = wave
        a = C1 * (y**3.)
    elif ispace==1:
        y = 1.0e4/wave
        a = C1 * (y**5.) / 1.0e4
    else:
        raise Exception('error in calc_planck: ISPACE must be either 0 or 1')

    tmp = C2 * y / T
    b = np.exp(tmp) - 1
    bb = (a/b)

    return bb


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Calculate mean molecular weight
"""
from nemesispy.common.info_mol import mol_info
from nemesispy.common.constants import AMU

def calc_mmw(ID, VMR, ISO=[]):
    """
    Calculate mean molecular weight in kg given a list of molecule IDs and
    a list of their respective volume mixing ratios.

    Parameters
    ----------
    ID : ndarray or list
        A list of Radtran gas identifiers.
    VMR : ndarray or list
        A list of VMRs corresponding to the gases in ID.
    ISO : ndarray or list
        If ISO = [] then terrestrial relative isotopic abundance is assumed
        for all gases.
        Otherwise, if ISO[i] = 0 then terrestrial relative isotopic abundance
        is assumed for the ith gas. To specify particular isotopologue, input
        the corresponding Radtran isotopologue identifiers.

    Returns
    -------
    mmw : real
        Mean molecular weight.
        Unit: kg

    Notes
    -----
    Cf mol_id.py and mol_info.py.
    """
    mmw = 0
    if len(ISO) == 0:
        for gas_index, gas_id in enumerate(ID):
            mmw += mol_info['{}'.format(gas_id)]['mmw']*VMR[gas_index]
    else:
        for gas_index, gas_id in enumerate(ID):
            if ISO[gas_index] == 0:
                mmw += mol_info['{}'.format(gas_id)]['mmw']*VMR[gas_index]
            else:
                mmw += mol_info['{}'.format(gas_id)]['isotope']\
                    ['{}'.format(ISO[gas_index])]['mass']*VMR[gas_index]
    mmw *= AMU # keep in SI unit
    return mmw

"""
import time
start = time.time()
for i in range(2000000):
    mmw = calc_mmw([1,2],VMR=[0.5,0.5],ISO=[1,1])
    # print('mmw',mmw)
end = time.time()
print('runtime1',end-start)

start = time.time()
for i in range(2000000):
    mmw = calc_mmw([1,2],VMR=[0.5,0.5],ISO=[0,0])
    # print('mmw',mmw)
end = time.time()
print('runtime2',end-start)

start = time.time()
for i in range(2000000):
    mmw = calc_mmw([1,2],VMR=[0.5,0.5])
    # print('mmw',mmw)
end = time.time()
print('runtime3',end-start)
"""

"""
import numpy as np
mmw = calc_mmw([1,2],VMR=[0.5,0.5],ISO=[1,1])
print('mmw',mmw)
mmw = calc_mmw([1,2],VMR=[0.5,0.5],ISO=[0,0])
print('mmw',mmw)
mmw = calc_mmw([1,2],VMR=[0.5,0.5])
print('mmw',mmw)

mmw = calc_mmw(np.array([1,2]),VMR=np.array([0.5,0.5]),ISO=np.array([1,1]))
print('mmw',mmw)
mmw = calc_mmw(np.array([1,2]),VMR=np.array([0.5,0.5]),ISO=np.array([0,0]))
print('mmw',mmw)
mmw = calc_mmw(np.array([1,2]),VMR=np.array([0.5,0.5]))
print('mmw',mmw)
"""

"""Test
# ID = [1,2,3]
# ISO = None
# VMR = [0.1,0.1,0.8]
# mmw = calc_mmw(ID,VMR,ISO)
ID = [1,2,3]
ISO = [1,1,1]
VMR = [0.1,0.1,0.8]
mmw = calc_mmw(ID,VMR,ISO)
ID = [1,2,3]
ISO = None
VMR = [0.1,0.1,0.8]
mmw = calc_mmw(ID,VMR,ISO)
"""


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Routines to calculate thermal emission spectra from a planetary atmosphere
using the correlated-k method to combine gaseous opacities.
We inlcude collision-induced absorption from H2-H2, H2-he, H2-N2, N2-Ch4, N2-N2,
Ch4-Ch4, H2-Ch4 pairs and Rayleigh scattering from H2 molecules and
He molecules.

As of now the routines are fully accelerated using numba.jit.
"""
import numpy as np
from numba import jit
from nemesispy.radtran.calc_planck import calc_planck
from nemesispy.radtran.calc_tau_gas import calc_tau_gas
from nemesispy.radtran.calc_tau_cia import calc_tau_cia
from nemesispy.radtran.calc_tau_rayleigh import calc_tau_rayleigh

@jit(nopython=True)
def calc_radiance(wave_grid, U_layer, P_layer, T_layer, VMR_layer, k_gas_w_g_p_t,
    P_grid, T_grid, del_g, ScalingFactor, RADIUS, solspec,
    k_cia, ID, cia_nu_grid, cia_T_grid, DEL_S):
    """
    Calculate emission spectrum using the correlated-k method.

    Parameters
    ----------
    wave_grid(NWAVE) : ndarray
        Wavelengths (um) grid for calculating spectra.
    U_layer(NLAYER) : ndarray
        Total number of gas particles in each layer.
        Unit: no. of particle/m^2
    P_layer(NLAYER) : ndarray
        Atmospheric pressure grid.
        Unit: Pa
    T_layer(NLAYER) : ndarray
        Atmospheric temperature grid.
        Unit: K
    VMR_layer(NLAYER,NGAS) : ndarray
        Array of volume mixing ratios for NGAS.
        Has dimensioin: NLAYER x NGAS
    k_gas_w_g_p_t(NGAS,NWAVE,NG,NPRESSK,NTEMPK) : ndarray
        k-coefficients.
        Has dimension: NGAS x NWAVE x NG x NPRESSK x NTEMPK.
    P_grid(NPRESSK) : ndarray
        Pressure grid on which the k-coeff's are pre-computed.
        We want SI unit (Pa) here.
    T_grid(NTEMPK) : ndarray
        Temperature grid on which the k-coeffs are pre-computed. In Kelvin
    del_g : ndarray
        Quadrature weights of the g-ordinates.
    ScalingFactor(NLAYER) : ndarray
        Scale stuff to line of sight
    RADIUS : real
        Planetary radius
        Unit: m
    solspec : ndarray
        Stellar spectra, used when the unit of the output is in fraction
        of stellar irradiance.

        Stellar flux at planet's distance (W cm-2 um-1 or W cm-2 (cm-1)-1)

    Returns
    -------
    spectrum : ndarray
        Output spectrum (W cm-2 um-1 sr-1)
    """
    # Reorder atmospheric layers from top to bottom
    ScalingFactor = ScalingFactor[::-1]
    P_layer = P_layer[::-1] # layer pressures (Pa)
    T_layer = T_layer[::-1] # layer temperatures (K)
    U_layer = U_layer[::-1] # layer absorber amounts (no./m^2)
    VMR_layer = VMR_layer[::-1,:] # layer volume mixing ratios
    DEL_S = DEL_S[::-1] # path lengths in each layer

    # Record array dimensions
    NGAS, NWAVE, NG, NPRESS, NTEMP = k_gas_w_g_p_t.shape
    NLAYER = len(P_layer)

    # Initiate arrays to record total optical paths
    tau_total_w_g_l = np.zeros((NWAVE,NG,NLAYER))

    # Collision induced absorptioin optical path (NWAVE x NLAYER)
    tau_cia = calc_tau_cia(wave_grid=wave_grid,K_CIA=k_cia,ISPACE=1,
        ID=ID,TOTAM=U_layer,T_layer=T_layer,P_layer=P_layer,VMR_layer=VMR_layer,
        DELH=DEL_S,cia_nu_grid=cia_nu_grid,TEMPS=cia_T_grid,INORMAL=1,NPAIR=9)

    # Rayleigh scattering optical path (NWAVE x NLAYER)
    tau_rayleigh = calc_tau_rayleigh(wave_grid=wave_grid,U_layer=U_layer)

    # Dust scattering optical path (NWAVE x NLAYER)
    tau_dust = np.zeros((NWAVE,NLAYER))

    # FORTRAN straight transcript
    tau_gas = calc_tau_gas(k_gas_w_g_p_t, P_layer, T_layer, VMR_layer, U_layer,
            P_grid, T_grid, del_g)

    # Merge all different opacities
    for ig in range(NG):
        tau_total_w_g_l[:,ig,:] = tau_gas[:,ig,:] + tau_cia[:,:] \
            + tau_dust[:,:] + tau_rayleigh[:,:]

    #Scale to the line-of-sight opacities
    tau_total_w_g_l = tau_total_w_g_l * ScalingFactor

    # Defining the units of the output spectrum / divide by stellar spectrum
    # IFORM = 1
    # radextra = np.sum(DEL_S[:-1])
    # radextra = 0
    # xfac = np.pi*4.*np.pi*((RADIUS+radextra)*1e2)**2./solspec[:]
    xfac = np.pi*4.*np.pi*(RADIUS*1e2)**2./solspec[:]

    # Calculating atmospheric gases contribution
    tau_cumulative_w_g = np.zeros((NWAVE,NG))
    tr_old_w_g = np.ones((NWAVE,NG))
    spec_w_g = np.zeros((NWAVE,NG))

    for ilayer in range(NLAYER):
        tau_cumulative_w_g[:,:] =  tau_total_w_g_l[:,:,ilayer] + tau_cumulative_w_g[:,:]
        tr_w_g = np.exp(-tau_cumulative_w_g[:,:]) # transmission function
        bb = calc_planck(wave_grid, T_layer[ilayer]) # blackbody function

        for iwave in range(NWAVE):
            for ig in range(NG):
                spec_w_g[iwave,ig] = spec_w_g[iwave,ig] \
                    + np.float32(tr_old_w_g[iwave,ig]-tr_w_g[iwave,ig])\
                    *bb[iwave]*xfac[iwave]

        tr_old_w_g = tr_w_g

    # surface/bottom layer contribution
    p1 = P_layer[int(len(P_layer)/2-1)] #midpoint
    p2 = P_layer[-1] # lowest point in altitude/highest in pressure

    surface = None
    if p2 > p1: # i.e. if not a limb path
        if not surface:
            radground = calc_planck(wave_grid,T_layer[-1])

        for ig in range(NG):
            spec_w_g[:,ig] = spec_w_g[:,ig] + tr_old_w_g[:,ig]*radground*xfac

    # Integrate over g-ordinates
    spectrum = np.zeros((NWAVE))
    for iwave in range(NWAVE):
        for ig in range(NG):
            spectrum[iwave] += spec_w_g[iwave,ig]*del_g[ig]

    return spectrum


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Interface class for running forward models.
"""
import numpy as np
from nemesispy.radtran.calc_mmw import calc_mmw
from nemesispy.radtran.calc_layer import calc_layer
from nemesispy.radtran.read import read_kls
from nemesispy.radtran.calc_radiance import calc_radiance
from nemesispy.radtran.read import read_cia
from nemesispy.common.calc_trig import gauss_lobatto_weights
from nemesispy.common.interpolate_gcm import interpvivien_point
from nemesispy.common.calc_hydrostat import calc_hydrostat

class ForwardModel():

    def __init__(self):
        """
        Attributes to store data that doesn't change during a retrieval
        """
        # planet and planetary system data
        self.M_plt = None
        self.R_plt = None
        self.M_star = None # currently not used
        self.R_star = None # currently not used
        self.T_star = None # currently not used
        self.semi_major_axis = None # currently not used
        self.NLAYER = None
        self.is_planet_model_set = False

        # opacity data
        self.gas_id_list = None
        self.iso_id_list = None
        self.wave_grid = None
        self.g_ord = None
        self.del_g = None
        self.k_table_P_grid = None
        self.k_table_T_grid = None
        self.k_gas_w_g_p_t = None
        self.cia_nu_grid = None
        self.cia_T_grid = None
        self.k_cia_pair_t_w = None
        self.is_opacity_data_set = False

        ### debug data
        # # layering debug data
        # self.U_layer = None
        # self.del_S = None
        # self.del_H = None
        # self.P_layer = None
        # self.T_layer = None
        # self.scale = None

        # phase curve debug data
        self.fov_H_model = None
        self.fake_fov_H_model = None
        fov_lattitudes = None
        fov_longitudes = None
        fov_emission_angles = None
        fov_weights = None
        self.total_weight = None

    def set_planet_model(self, M_plt, R_plt, gas_id_list, iso_id_list, NLAYER,
        gas_name_list=None, solspec=None, R_star=None, T_star=None,
        semi_major_axis=None):
        """
        Store the planetary system parameters
        """
        self.M_plt = M_plt
        self.R_plt = R_plt
        # self.R_star = R_star
        # self.T_star = T_star
        # self.semi_major_axis = semi_major_axis
        self.gas_name_list = gas_name_list
        self.gas_id_list = gas_id_list
        self.iso_id_list = iso_id_list
        self.NLAYER = NLAYER

        self.is_planet_model_set = True

    def set_opacity_data(self, kta_file_paths, cia_file_path):
        """
        Read gas ktables and cia opacity files and store as class attributes.
        """
        k_gas_id_list, k_iso_id_list, wave_grid, g_ord, del_g, k_table_P_grid,\
            k_table_T_grid, k_gas_w_g_p_t = read_kls(kta_file_paths)
        """
        Some gases (e.g. H2 and He) have no k table data so gas id lists need
        to be passed somewhere else.
        """
        self.k_gas_id_list = k_gas_id_list
        self.k_iso_id_list = k_iso_id_list
        self.wave_grid = wave_grid
        self.g_ord = g_ord
        self.del_g = del_g
        self.k_table_P_grid = k_table_P_grid
        self.k_table_T_grid = k_table_T_grid
        self.k_gas_w_g_p_t = k_gas_w_g_p_t # key

        cia_nu_grid, cia_T_grid, k_cia_pair_t_w = read_cia(cia_file_path)
        self.cia_nu_grid = cia_nu_grid
        self.cia_T_grid = cia_T_grid
        self.k_cia_pair_t_w = k_cia_pair_t_w

        self.is_opacity_data_set = True

    def calc_point_spectrum(self, H_model, P_model, T_model, VMR_model,
        path_angle, solspec=[]):
        """
        First get layer properties from model inputs
        Then calculate the spectrum at a single point on the disc.
        """
        H_layer,P_layer,T_layer,VMR_layer,U_layer,\
        MMW_layer,Gas_layer,scale,del_S,del_H = calc_layer(
            self.R_plt, H_model, P_model, T_model, VMR_model,
            self.gas_id_list, self.NLAYER, path_angle, layer_type=1,
            H_0=0.0, NSIMPS=101
            )

        if len(solspec)==0:
            solspec = np.ones(len(self.wave_grid))

        point_spectrum = calc_radiance(self.wave_grid, U_layer, P_layer, T_layer,
            VMR_layer, self.k_gas_w_g_p_t, self.k_table_P_grid,
            self.k_table_T_grid, self.del_g, ScalingFactor=scale,
            RADIUS=self.R_plt, solspec=solspec, k_cia=self.k_cia_pair_t_w,
            ID=self.gas_id_list,cia_nu_grid=self.cia_nu_grid,
            cia_T_grid=self.cia_T_grid, DEL_S=del_H)

        return point_spectrum

    def calc_point_spectrum_hydro(self, P_model, T_model, VMR_model,
        path_angle, solspec=[]):
        """
        Use the hydrodynamic equation to calculate layer height
        First get layer properties from model inputs
        Then calculate the spectrum at a single point on the disc.
        """
        NPRO = len(P_model)
        mmw = np.zeros(NPRO)
        for ipro in range(NPRO):
            mmw[ipro] = calc_mmw(self.gas_id_list,VMR_model[ipro,:])
        H_model = calc_hydrostat(P=P_model, T=T_model, mmw=mmw,
            M_plt=self.M_plt, R_plt=self.R_plt)

        H_layer,P_layer,T_layer,VMR_layer,U_layer,\
        MMW_layer,Gas_layer,scale,del_S,del_H = calc_layer(
            self.R_plt, H_model, P_model, T_model, VMR_model,
            self.gas_id_list, self.NLAYER, path_angle, layer_type=1,
            H_0=0.0, NSIMPS=101
            )

        if len(solspec)==0:
            solspec = np.ones(len(self.wave_grid))

        point_spectrum = calc_radiance(self.wave_grid, U_layer, P_layer, T_layer,
            VMR_layer, self.k_gas_w_g_p_t, self.k_table_P_grid,
            self.k_table_T_grid, self.del_g, ScalingFactor=scale,
            RADIUS=self.R_plt, solspec=solspec, k_cia=self.k_cia_pair_t_w,
            ID=self.gas_id_list,cia_nu_grid=self.cia_nu_grid,
            cia_T_grid=self.cia_T_grid, DEL_S=del_H)

        return point_spectrum

    def calc_disc_spectrum(self,phase,nmu,P_model,
        global_model_P_grid,global_T_model,global_VMR_model,
        mod_lon,mod_lat,solspec):
        """
        Parameters
        ----------
        phase : real
            Orbital phase, increase from 0 at primary transit to 180 and secondary
            eclipse.

        """
        # initialise output array
        disc_spectrum = np.zeros(len(self.wave_grid))

        # get locations and angles for disc averaging
        nav, wav = gauss_lobatto_weights(phase, nmu)
        wav = np.around(wav,decimals=8)
        fov_lattitudes = wav[0,:]
        fov_longitudes = wav[1,:]
        fov_stellar_zen = wav[2,:]
        fov_emission_angles = wav[3,:]
        fov_stellar_azi = wav[4,:]
        fov_weights = wav[5,:]

        for iav in range(nav):
            xlon = fov_longitudes[iav]
            xlat = fov_lattitudes[iav]
            T_model, VMR_model = interpvivien_point(
                lon=xlon,lat=xlat, p=P_model,
                gcm_lon=mod_lon, gcm_lat=mod_lat,
                gcm_p=global_model_P_grid,
                gcm_t=global_T_model, gcm_vmr=global_VMR_model,
                substellar_point_longitude_shift=180)
            path_angle = fov_emission_angles[iav]
            weight = fov_weights[iav]
            NPRO = len(P_model)
            mmw = np.zeros(NPRO)
            for ipro in range(NPRO):
                mmw[ipro] = calc_mmw(self.gas_id_list,VMR_model[ipro,:])
            H_model = calc_hydrostat(P=P_model, T=T_model, mmw=mmw,
                M_plt=self.M_plt, R_plt=self.R_plt)
            point_spectrum = self.calc_point_spectrum(
                H_model, P_model, T_model, VMR_model, path_angle,
                solspec=solspec)
            disc_spectrum += point_spectrum * weight
        # print('number of point spectrum = ',nav)
        return disc_spectrum

    def calc_disc_spectrum_uniform(self, nmu, P_model, T_model, VMR_model,
        H_model=[],solspec=[]):
        """Caculate the disc integrated spectrum of a homogeneous atmosphere
        """
        # initialise output array
        disc_spectrum = np.zeros(len(self.wave_grid))
        nav, wav = gauss_lobatto_weights(0, nmu)
        fov_emission_angles = wav[3,:]
        fov_weights = wav[5,:]

        # Hydrostatic case
        if len(H_model) == 0:
            for iav in range(nav):
                path_angle = fov_emission_angles[iav]
                weight = fov_weights[iav]
                point_spectrum = self.calc_point_spectrum_hydro(
                    P_model, T_model, VMR_model, path_angle,
                    solspec=solspec)
                disc_spectrum += point_spectrum * weight
        else:
            for iav in range(nav):
                path_angle = fov_emission_angles[iav]
                weight = fov_weights[iav]
                point_spectrum = self.calc_point_spectrum(
                    H_model, P_model, T_model, VMR_model, path_angle,
                    solspec=solspec)
                disc_spectrum += point_spectrum * weight
        return disc_spectrum


from setuptools import setup, find_packages

VERSION = "0.0.1"
DESCRIPTION = "Tools for modelling exoplanet spectra"
setup(name="nemesispy",
      version=VERSION,
      description=DESCRIPTION,
      url='https://github.com/Jingxuan97/nemesispy2022',
      author="Jingxuan Yang",
      author_email="jingxuanyang15@gmail.com",
      packages=find_packages(),
      install_requires=[
          "numpy",
          "scipy",
          "matplotlib",
          "numba"]
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Generate sample locations and corresponding weights on a planetary disc
for calculating the disc-averaged radiance of a transitting planet
at variable orbital phases.

Solar zenith and azimuth angles of the sample points
are also generated, but not currently used asreflected sunlight is not included.

Assumptions and conventions:

    1. The visible planetary disc is a circle (with dimensionless radius 1).

    2. The orbital phase is 0 at primary transit and 180 at secondary eclipse.
    Orbital phase increases in the direction of orbital motion.

    3. The planet is tidally locked, i.e. it's in synchronous rotation
    and always present the same hemisphere to the star. Therefore, We are
    observing the planet's orbit edge on.

    4. The planetocentric longitude is defined such that the antistellar point
    is on the 0E meridian, and the substellar point is on the 180E meridian.
    Longitude increases in the direction of planet's self-rotation.

    5. The north pole of the planet is defined such that when viewed directly
    above the north pole the planet is rotating anticlockwise.

Coordinate system (origin is at the centre of the target):

    x-axis points from observer's 9 o'clock to observer's 3 o'clock
    y-axis points from observer to the target
    z-axis points from south pole to north pole
    r is the radial distance
    theta is measured clockwiseky from the z-axis
    phi is measured anticlockwise from x-axis

    x = r * sin(theta) * cos(phi) = rho * cos(alpha)
    y = r * sin(theta) * sin(phi) = rho * sin(alpha)
    z = r * cos(theta)

    rho: projected distance from a point on the disc to the centre of the disc
    alpha: argument of a point on the projected disc, measured anticlockwise from x axis
"""
import numpy as np
from numba import jit

@jit(nopython=True)
def arctan(x,y):
    """
    Calculate the argument of the point (x,y) in the range [0,2pi).

    Parameters
    ----------
        x : real
            x-coordinate of the point (length of the adjacent side)
        y : real
            y-coordinate of the point (length of the opposite side)
    Returns
    -------
        ang : real
            Argument of (x,y) in radians
    """
    if(x == 0.0):
        if (y == 0.0) : ang = 0.0 # (x,y) is the origin, ill-defined
        elif (y > 0.0) : ang = 0.5*np.pi # (x,y) is on positive y-axis
        else : ang = 1.5*np.pi  # (x,y) is on negative y-axis
    else:
        ang=np.arctan(y/x)
        if (y > 0.0) :
            if (x > 0.0) : ang = ang # (x,y) is in 1st quadrant
            else : ang = ang+np.pi # (x,y) is in 2nd quadrant
        elif (y == 0.0) :
            if (x > 0.0) : ang = 0 # (x,y) is on positive x-axis
            else : ang = np.pi # (x,y) is on negative x-axis
        else:
            if (x > 0.0) : ang = ang+2*np.pi # (x,y) is in 4th quadrant
            else : ang = ang+np.pi # (x,y) is in 3rd quadrant
    return ang

@jit(nopython=True)
def rotatey(v, phi):
    """
    Rotate a 3D vector v anticlockwisely about the y-axis by angle phi

    Parameters
    ----------
        v : ndarray
            A real 3D vector to rotate
        phi : real
            Angle to rotate the vector v by (radians)

    Returns
    -------
        v_new : ndarray
            Rotated 3D vector
    """
    a = np.zeros((3,3)) # construct the rotation matrix
    a[0,0] = np.cos(phi)
    a[0,2] = np.sin(phi)
    a[1,1] = 1.
    a[2,0] = -np.sin(phi)
    a[2,2] = np.cos(phi)
    # v_new = np.matmul(a,v) # unsupported NumPy function
    v_new = np.zeros(3)
    for i in range(3):
        for j in range(3):
            v_new[i] += a[i,j] * v[j]
    return v_new

@jit(nopython=True)
def rotatez(v, phi):
    """
    Rotate a 3D vector v anticlockwisely about the z-axis by angle phi

    Parameters
    ----------
        v : ndarray
            A real 3D vector to rotate
        phi : real
            Angle to rotate the vector by (radians)

    Returns
    -------
        v_new : ndarray
            Rotated 3D vector
    """
    a = np.zeros((3,3))
    a[0,0] = np.cos(phi)
    a[0,1] = -np.sin(phi)
    a[1,0] = np.sin(phi)
    a[1,1] = np.cos(phi)
    a[2,2] = 1
    # v_new = np.matmul(a,v) # unsupported NumPy function
    v_new = np.zeros(3)
    for i in range(3):
        for j in range(3):
            v_new[i] += a[i,j] *v[j]
    return v_new

@jit(nopython=True)
def generate_angles(phase,rho,alpha):
    """
    Finds the stellar zenith angle, stellar azimuth angle, lattitude and longitude
    of a chosen point on the visible disc of a planet under observation. The planet
    is assumed to be tidally locked, and is observed on an edgy-on orbit.

    Refer to the begining of the trig.py file for geomety and convections.

    Parameters
    ----------
    phase : real
        Orbital phase in degrees. 0 at  parimary transit and 180 at secondary eclipse.
        Range: [0,360)
    rho	: real
        Fractional radius of the point on disc, must be between 0 and 1 inclusive.
        Range: [0,1]
    alpha : real
        Argument of the point on visible disc (degrees), measured
        anticlockwise from 3 o'clock.
        Range: [0,360)

    Returns
    -------
    zen : real
        Computed solar zenith angle (radians), which is the angle between the local
        normal and the stellar direction vector.
    azi	: real
        Computed solar azimuth angle (radians). Uses convention that
        forward scatter = 0. !!! need to define.
    lat	: real
        Planetocentric latitude of the point (degrees).
    lon	: real
        Planetocentric longitude of the point (degrees).
    """
    phase = np.mod(phase,360)
    dtr = np.pi/180. # degree to radiance conversion factor
    assert rho <=1, "Fractional radius should be less or equal to 1"

    # get stellar direction vector in Cartesian coordinates
    # ie unit vector in direction of star
    theta_star = np.pi/2. # star lies in planet's equitorial plane
    phi_star = 90.0 + phase # when phase angle is 0 the star lies on the y-axis
    x_star = np.sin(theta_star)*np.cos(phi_star*dtr)
    y_star = np.sin(theta_star)*np.sin(phi_star*dtr)
    z_star = np.cos(theta_star)
    v_star = np.array([x_star,y_star,z_star])

    # get Cartesian coordinates of input point
    # calculate point position vector using spherical polars (r=1,theta,phi)
    theta_point = np.arccos(rho*np.sin(alpha*dtr)) # ie planetary zenith angle of poiny
    if np.sin(theta_point) != 0.0:
        cos_phi = rho*np.cos(alpha*dtr)/abs(np.sin(theta_point))
        phi_point = (-np.arccos(cos_phi))%(2*np.pi) # azimuth angle of point (on our side)
    else:
        phi_point = 0.0 # sin(theta_point) = 0 at north polt
    x_point = np.sin(theta_point)*np.cos(phi_point)
    y_point = np.sin(theta_point)*np.sin(phi_point)
    z_point = np.cos(theta_point)
    v_point = np.array([x_point,y_point,z_point])

    # calculate angle between solar position vector and local normal
    # i.e. zen solar zenith angle
    inner_product = np.sum(v_star*v_point)
    zen = np.arccos(inner_product)
    zen = np.around(zen, 10)

    # calculate latitude and longitude of the spot
    # (sub-stellar point = 180E, anti-stellar point = 0E, longtitudes in the direction of self-rotation)
    lat = np.around(90.-theta_point*180/np.pi, 10) # southern hemisphere has negative lattitude
    lon = (phi_point/dtr - (phi_star+180))%360 # substellar point is 180E

    # calculate emission viewing angle direction vector (-y axis) (Observer direction vecto)
    x_observer = 0.
    y_observer = -1.0
    z_observer = 0.0
    v_observer = np.array([x_observer,y_observer,z_observer])

    ### calculate azimuth angle
    # Rotate frame clockwise by phi_point about z (v_point is now x-axis)
    v_star_1=rotatez(v_star,-phi_point)
    v_point_1=rotatez(v_point,-phi_point)
    v_observer_1=rotatez(v_observer,-phi_point)

    # Rotate frame clockwise by theta_point about y (v_point is now z-axis )
    v1B=rotatey(v_star_1,-theta_point)
    v2B=rotatey(v_point_1,-theta_point)
    v3B=rotatey(v_observer_1,-theta_point)

    # thetsolB=np.arccos(v1B[2])
    # thetobsB=np.arccos(v3B[2])
    phisolB=arctan(v1B[0], v1B[1])
    phiobsB=arctan(v3B[0], v3B[1])

    azi = abs(phiobsB-phisolB)
    if(azi > np.pi):
        azi=2*np.pi-azi

    # Ensure azi meets convention where azi=0 means forward-scattering
    azi = np.pi-azi

    return zen, azi, lat, lon

@jit(nopython=True)
def gauss_lobatto_weights(phase, nmu):
    """
    Given the orbital phase, calculates the coordinates and weights of the points
    on a disc needed to compute the disc integrated radiance.

    The points are chosen on a number of rings according to Gauss-Lobatto quadrature
    scheme, and spaced on the rings according to trapezium rule.

    Refer to the begining of the trig.py file for geomety and convections.

    Parameters
    ----------
    phase : real
        Stellar phase/orbital phase in degrees.
        0=parimary transit and increase to 180 at secondary eclipse.
    nmu	: integer
        Number of zenith angle ordinates

    Output variables
    nav	: integer
        Number of FOV points
    wav	: ndarray
        FOV-averaging table:
        0th row is lattitude, 1st row is longitude, 2nd row is stellar zenith
        angle, 3rd row is emission zenith angle, 4th row is stellar azimuth angle,
        5th row is weight.
    """
    assert nmu <=5, "Currently cannot do more than 5 quadrature rings"
    assert nmu >=2, "Need at least 2 quadrature rings"
    phase = phase%360
    dtr = np.pi/180 # degree to radiance conversion factor
    delR = 1./nmu

    # set up the output arrays
    nsample = 1000 # large array size to hold calculations
    tablat = np.zeros(nsample) # latitudes
    tablon = np.zeros(nsample) # longitudeds
    tabzen = np.zeros(nsample) # zenith angle in quadrature scheme
    tabsol = np.zeros(nsample) # solar zenith angle
    tabazi = np.zeros(nsample) # solar azimuth angle (scattering phase angle?)
    tabwt = np.zeros(nsample)  # weight of each sample

    # list the gauss labatto quadrature points and weights
    if nmu == 2:
        mu = [0.447213595499958,1.000000]                   # cos zenith angle
        wtmu = [0.8333333333333333,0.166666666666666666]    # corresponding weights
    if nmu == 3:
        mu = [0.28523151648064509,0.7650553239294646,1.0000]
        wtmu = [0.5548583770354863,0.3784749562978469,0.06666666666666666]
    if nmu == 4:
        mu = [0.2092992179024788,0.5917001814331423,0.8717401485096066,
              1.00000]
        wtmu = [0.4124587946587038,0.3411226924835043,0.2107042271435060,
                0.035714285714285]
    if nmu == 5:
        mu = [0.165278957666387,0.477924949810444,0.738773865105505,
              0.919533908166459,1.00000000000000]
        wtmu = [0.327539761183898,0.292042683679684,0.224889342063117,
                0.133305990851069,2.222222222222220E-002]


    # # define a range of angles from 0 to 360
    # """
    # thet = np.arange(361)
    # """

    # """
    # xx = np.around(np.cos(thet*dtr), 14)
    # zz = np.around(np.sin(thet*dtr), 14)
    # """

    # trace out the day/night terminator
    z_term = np.linspace(-1,1,201) # pick out some z coordinates for the terminator
    if 0<= phase <= 180:
        # terminator is to the left of z axis
        theta_term = 2*np.pi - np.arccos(z_term)
    else:
        # terminator in the right side of the disc
        theta_term = np.arccos(z_term)
    x_term = np.sin(theta_term) * np.around(np.cos(phase*dtr),14) # x coords of terminator
    r_term = np.sqrt(x_term**2+z_term**2) # radial coords of terminator
    rmin = min(r_term) # least radius (is on x axis)

    # define FOV averaging points
    isample = 0
    for imu in range(0, nmu): # quadrature rings
        r_quad = np.sqrt(1.-mu[imu]**2) # quadrature radius (from small to large)
        half_circum = np.pi*r_quad # half the circumference

        # see if the quadrature ring intersects the terminator
        # if so, find the intersection point and place a sample point there
        if r_quad > rmin: # quadrature ring intersects the terminator
            ikeep = np.where(r_term<=r_quad)
            ikeep = ikeep[0] # index of the points on the terminator with radius > r_quad
            i_intersect = np.array([ikeep[0], ikeep[-1]]) # index of two intersectionns
            x_intersect = x_term[i_intersect] # x coordinates of intersection
            z_intersect = z_term[i_intersect] # z coordinates of intersection

            # take the intersection in the upper hemisphere
            if z_intersect[1] > 0:
                alpha_intersect = arctan(x_intersect[1],z_intersect[1])/dtr
            else:
                alpha_intersect = arctan(x_intersect[0],z_intersect[0])/dtr

            # place the sample points on the quadrature rings on either side of the intersection
            nalpha1 = int(0.5+half_circum*(alpha_intersect/180.0)/delR) # round up; separation ~ R/nmu
            nalpha2 = int(0.5+half_circum*((180.-alpha_intersect)/180.0)/delR)

            # at least 1 point either side of the intersection
            if(nalpha1 < 2):
                nalpha1=2
            if(nalpha2 < 2):
                nalpha2=2

            # set the alphas of the sample points on current quadrature ring
            nalpha = nalpha1+nalpha2-1 # intersection point double counted
            alpha1 = alpha_intersect/(nalpha1-1) * np.arange(nalpha1)
            alpha2 = alpha_intersect+(180.-alpha_intersect)/(nalpha2-1) * np.arange(nalpha2)
            alpha2 = alpha2[1:(nalpha2)] # intersect was counted twice
            alpha_sample_list = np.concatenate((alpha1,alpha2))

        else: # quadrature ring does not intersect terminator
            if(half_circum > 0.0):
                nalpha = int(0.5+half_circum/delR)
                alpha_sample_list = 180*np.arange(nalpha)/(nalpha-1)
            else:
                nalpha=1

        if(nalpha > 1): # more than one sample on the quadrature ring

            for ialpha in np.arange(0,nalpha):

                alpha_sample = alpha_sample_list[ialpha]

                thetasol_sample, azi_sample, lat_sample, lon_sample \
                    = generate_angles(phase,r_quad,alpha_sample)

                # trapezium rule weights
                if (ialpha == 0):
                    wt_trap = (alpha_sample_list[ialpha+1]-alpha_sample_list[ialpha])/2.0
                elif (ialpha == nalpha-1):
                    wt_trap = (alpha_sample_list[ialpha]-alpha_sample_list[ialpha-1])/2.0
                else:
                    wt_trap = (alpha_sample_list[ialpha+1]-alpha_sample_list[ialpha-1])/2.0

                wt_azi= wt_trap/180. # sample azimuthal weight


                tablat[isample] = lat_sample # sample lattitude
                tablon[isample] = lon_sample # sample longitude
                tabzen[isample] = np.arccos(mu[imu])/dtr # sample emission zenith angle
                tabsol[isample] = thetasol_sample/dtr # sample stellar zenith angle
                tabazi[isample] = azi_sample/dtr # sample stellar azimuth angle
                tabwt[isample] = 2*mu[imu]*wtmu[imu]*wt_azi # sample weight
                isample = isample+1

        else:
            alpha_sample = 0.
            thetasol_sample,azi_sample, lat_sample,lon_sample \
                = generate_angles(phase,r_quad,alpha_sample)
            if(tabzen[isample] == 0.0):
                azi_sample = 180.
            tablat[isample] = lat_sample
            tablon[isample] = lon_sample
            tabzen[isample] = np.arccos(mu[imu])/dtr
            tabsol[isample] = thetasol_sample/dtr
            tabazi[isample] = azi_sample
            tabwt[isample] = 2*mu[imu]*wtmu[imu]
            isample = isample+1

    nav = isample
    wav = np.zeros((6,isample))
    sum=0.
    for i in np.arange(0,isample):
        wav[0,i]=tablat[i]              # 0th array is lattitude
        wav[1,i]=tablon[i]%360          # 1st array is longitude
        wav[2,i]=tabsol[i]              # 2nd array is stellar zenith angle
        wav[3,i]=tabzen[i]              # 3rd array is emission zenith angle
        wav[4,i]=tabazi[i]              # 4th array is stellar azimuth angle
        wav[5,i]=tabwt[i]               # 5th array is weight
        sum = sum+tabwt[i]

    for i in range(isample):            # normalise weights so they add up to 1
        wav[5,i]=wav[5,i]/sum

    # wav = np.around(wav,decimals=8)

    return nav, wav



# -*- coding: utf-8 -*-
from scipy.special import erf
from scipy.special import voigt_profile
import numpy as np

def constant(x, C):
    return C * x/x

def lorentz(x, A, mu, gamma):
    return A * (gamma**2/( (x-mu)**2 + gamma**2))

def lorentz_plus_C(x, A, mu, gamma, C):
    out = lorentz(x, A, mu, gamma) + C
    return out

def normal(x, A, mu, sigma):
    out = A * 1/(sigma*(2*np.pi)**0.5) * np.exp(-0.5*(x-mu)**2/sigma**2)
    return out

def normal_plus_C(x, A, mu, sigma, C):
    out = normal(x, A, mu, sigma) + C
    return out

def skew_normal(x, A, mu, sigma, s):
    phi =  normal(x, A, mu, sigma)
    PHI = 0.5 * (1 + erf( s*((x-mu)/sigma) ))
    out = phi * PHI
    return out

def voigt(x, A, mu, sigma, gamma):
    return A*voigt_profile(x-mu,sigma,gamma)

def voigt_plus_C(x, A, mu, sigma, gamma, C):
    out = voigt(x, A, mu, sigma, gamma) + C
    return out

def skew_voigt(x, A, mu, sigma, gamma, skew):
    v = voigt(x, A, mu, sigma, gamma)
    PHI = 0.5 * (1 + erf( skew*((x-mu)/sigma) ))
    out = v * PHI
    return out

def fourier1(x, c0, c1, s1):
    x = x/180*np.pi
    sin = s1 * np.sin(x)
    cos = c1 * np.cos(x)
    return cos + sin + c0

def fourier2(x, c0, c1, c2, s1, s2):
    x = x/180*np.pi
    cos = c1 * np.cos(x) + c2 * np.cos(2*x)
    sin = s1 * np.sin(x) + s2 * np.sin(2*x)
    return cos + sin + c0

def fourier3(x, c0, c1, c2, c3, s1, s2, s3):
    x = x/180*np.pi
    cos = c1 * np.cos(x) + c2 * np.cos(2*x) + c3 * np.cos(3*x)
    sin = s1 * np.sin(x) + s2 * np.sin(2*x) + s3 * np.sin(3*x)
    return cos + sin + c0

def fourier4(x, c0, c1, c2, c3, c4, s1, s2, s3, s4):
    x = x/180*np.pi
    cos = c1 * np.cos(x) + c2 * np.cos(2*x) + c3 * np.cos(3*x) + c4 * np.cos(4*x)
    sin = s1 * np.sin(x) + s2 * np.sin(2*x) + s3 * np.sin(3*x) + s4 * np.sin(4*x)
    return cos + sin + c0
